
from epyk.core.html import Html
from epyk.core.html.options import OptChartJs

from epyk.core.js import JsUtils
from epyk.core.js.packages import packageImport
from epyk.core.js.primitives import JsObject

from epyk.core.js.packages import JsChartJs
from epyk.core.js.packages import JsD3


class Chart(Html.Html):
  name = 'ChartJs Chart'
  requirements = ('Chart.js', )

  def __init__(self,  report, width, height, htmlCode, options, profile):
    self.height = height[0]
    super(Chart, self).__init__(report, [], htmlCode=htmlCode, css_attrs={"width": width, "height": height}, profile=profile)
    self._d3, self._chart, self._datasets, self._options, self._data_attrs, self._attrs = None, None, [], None, {}, {}
    self._options_init = options
    self.style.css.margin_top = 10

  @property
  def chartId(self):
    """
    Description:
    -----------
    Return the Javascript variable of the chart
    """
    return "%s_obj" % self.htmlCode

  @property
  def d3(self):
    if self._d3 is None:
      self._d3 = JsD3.D3Select(self._report, selector="d3.select('#%s')" % self.htmlCode, setVar=False)
    return self._d3

  @property
  def js(self):
    """
    Description:
    -----------
    Javascript base function

    Return all the Javascript functions defined in the framework.
    THis is an entry point to the full Javascript ecosystem.

    :return: A Javascript object

    :rtype: JsChartJs.JsChart
    """
    if self._js is None:
      self._js = JsChartJs.ChartJs(selector="window['%s']" % self.chartId, src=self)
    return self._js

  @property
  def options(self):
    """
    Description:
    -----------

    :rtype: OptChartJs.Options
    """
    if self._options is None:
      self._options = OptChartJs.Options(self._report, attrs=self._options_init)
    return self._options

  @property
  def plugins(self):
    return self.options.plugins

  @packageImport('chartjs-plugin-dragdata')
  def dragData(self):
    """
    Description:
    -----------
    A plugin for Chart.js >= 2.4.0

    Makes data points draggable. Supports touch events.

    Related Pages:

      https://github.com/chrispahm/chartjs-plugin-dragdata
    """
    self.options._attrs['dragData'] = True
    return self.options

  def labels(self, labels):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param labels:
    """
    self._data_attrs['labels'] = labels
    return self

  def label(self, i, name):
    """
    Description:
    -----------
    Change the series name

    Attributes:
    ----------
    :param i: Integer. The series index according to the y_columns
    :param name: String. The new name to be set
    """
    self.dataset(i).label = name
    return self

  def dataset(self, i=None):
    """
    Description:
    -----------

    :rtype: JsChartJs.DataSetPie
    """
    if i is None:
      return self._datasets[-1]

    return self._datasets[i]

  def click(self, jsFncs, profile=False):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsFncs:
    :param profile:
    """
    if self._attrs.get('type') in ['pie']:
      tmpJsFncs = ["var activePoints = %s.getSegmentsAtEvent(event)" % self.chartId]
      tmpJsFncs.append("if(activePoints.length > 0){ %s }" % JsUtils.jsConvertFncs(jsFncs, toStr=True))
    else:
      tmpJsFncs = ["var activePoints = %s.getElementsAtEvent(event)" % self.chartId]
      tmpJsFncs.append("if(activePoints.length > 0){ %s }" % JsUtils.jsConvertFncs(jsFncs, toStr=True))
    return super(Chart, self).click(tmpJsFncs, profile)

  def hover(self, jsFncs, profile=False):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param jsFncs:
    :param profile:
    """
    tmpJsFncs = ["var activePoints = %s.getElementsAtEvent(event)" % self.chartId]
    tmpJsFncs.append("if(activePoints.length > 0){ %s }" % JsUtils.jsConvertFncs(jsFncs, toStr=True))
    return self.on("mouseover", tmpJsFncs, profile)

  @property
  def datasets(self):
    return self._datasets

  def getCtx(self):
    obj_datasets = "[%s]" % ", ".join([d.toStr() for d in self._datasets])
    self._data_attrs['datasets'] = JsObject.JsObject.get(obj_datasets)
    obj_data = "{%s}" % ", ".join(["%s: %s" % (k, JsUtils.jsConvertData(v, None)) for k, v in self._data_attrs.items()])
    self._attrs["data"] = JsObject.JsObject.get(obj_data)
    self._attrs["options"] = JsObject.JsObject.get(str(self.options))
    str_ctx = "{%s}" % ", ".join(["%s: %s" % (k, JsUtils.jsConvertData(v, None)) for k, v in self._attrs.items()])
    return str_ctx

  def convert(self, data, options, profile=False):
    mod_name = __name__.split(".")[-1]
    constructors = self._report._props.setdefault("js", {}).setdefault("constructors", {})
    constructors[self.builder_name] = "function %s%sConvert(data, options){%s; return result}" % (
      mod_name, self.builder_name, self._js__convertor__)
    if isinstance(data, dict):
      # check if there is no nested HTML components in the data
      tmp_data = ["%s: %s" % (JsUtils.jsConvertData(k, None), JsUtils.jsConvertData(v, None)) for k, v in data.items()]
      js_data = "{%s}" % ",".join(tmp_data)
    else:
      js_data = JsUtils.jsConvertData(data, None)
    options, js_options = options or self._options_init, []
    for k, v in options.items():
      if isinstance(v, dict):
        row = ["'%s': %s" % (s_k, JsUtils.jsConvertData(s_v, None)) for s_k, s_v in v.items()]
        js_options.append("'%s': {%s}" % (k, ", ".join(row)))
      else:
        if str(v).strip().startswith("function"):
          js_options.append("%s: %s" % (k, v))
        else:
          js_options.append("%s: %s" % (k, JsUtils.jsConvertData(v, None)))
    return "%s%sConvert(%s, %s)" % (mod_name, self.builder_name, js_data, "{%s}" % ",".join(js_options))

  def build(self, data=None, options=None, profile=False):
    if data:
      return "Object.assign(window['%(chartId)s'].data, %(data)s); window['%(chartId)s'].update()" % {'chartId': self.chartId, 'data': self.convert(data, options, profile)}
      #return "window['%(chartId)s'].data.labels = %(labels)s; window['%(chartId)s'].data.datasets = [%(recordsset)s]; window['%(chartId)s'].update()" % {'labels': js_data['labels'], 'recordsset': ",".join(recordsset),  'chartId': self.chartId}

    return '%s = new Chart(%s.getContext("2d"), %s)' % (self.chartId, self.dom.varId, self.getCtx())

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<canvas %s></canvas>' % self.get_attrs(pyClassNames=self.style.get_classes())


class Datasets(object):

  def __init__(self, report):
    self._report, self.__data = report, []

  def add(self, data):
    self.__data.append(data)
    return self


class ChartLine(Chart):

  def __init__(self, report, width, height, htmlCode, options, profile):
    super(ChartLine, self).__init__(report, width, height, htmlCode, options, profile)
    self._attrs['type'] = 'line'

  @property
  def options(self):
    """
    Description:
    ------------

    :rtype: OptionsPolar
    """
    if self._options is None:
      self._options = OptChartJs.OptionsLine(self._report, attrs=self._options_init)
    return self._options

  def new_dataset(self, id, data, label, colors=None, opacity=0.8, type=None):
    data = JsChartJs.DataSetScatterLine(self._report, attrs={"data": data})
    data.fill = False
    data.label = label
    if colors is None:
      data.borderColor = self._report.theme.charts[id]
      data.backgroundColor = self._report.theme.charts[id]
    data.borderWidth = 1
    data.pointRadius = 3
    if opacity is not None:
      data.fillOpacity = opacity
    return data

  def add_dataset(self, data, label="", colors=None, opacity=None):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param data:
    :param label:
    :param colors:
    :param opacity:
    """
    data = self.new_dataset(len(self._datasets), data, label, colors=colors, opacity=opacity, type=None)
    self._datasets.append(data)
    return data

  @property
  def _js__convertor__(self):
    return ''' 
      if(data.python){
          result = {datasets: [], labels: data.series};
          data.datasets.forEach(function(rec, i){
            result.datasets.push( {label: data.series[i], data: rec, backgroundColor: options.colors[i], borderColor: options.colors[i]} )
          })}
      else{
        var temp = {}; var labels = []; var uniqLabels = {};
        options.y_columns.forEach(function(series){temp[series] = {}});
        data.forEach(function(rec){ 
          options.y_columns.forEach(function(name){
            if(rec[name] !== undefined){
              if (!(rec[options.x_column] in uniqLabels)){labels.push(rec[options.x_column]); uniqLabels[rec[options.x_column]] = true};
              temp[name][rec[options.x_column]] = rec[name]}})});
        result = {datasets: [], labels: labels};
        options.y_columns.forEach(function(series, i){
          dataSet = {label: series, data: [], backgroundColor: options.colors[i], borderColor: options.colors[i]};
          for(var attr in options.attrs){dataSet[attr] = options.attrs[attr]};
          labels.forEach(function(x){
            if (temp[series][x] == undefined) {dataSet.data.push(null)} else{dataSet.data.push(temp[series][x])}
          }); result.datasets.push(dataSet)})
      }'''


class ChartBubble(Chart):

  def __init__(self, report, width, height, htmlCode, options, profile):
    super(ChartBubble, self).__init__(report, width, height, htmlCode, options, profile)
    self._attrs['type'] = 'bubble'

  def new_dataset(self, id, data, label, colors=None, opacity=0.8, type=None):
    data = JsChartJs.DataSetBubble(self._report, attrs={"data": data})
    data.fill = False
    data.label = label
    if colors is None:
      data.borderColor = self._report.theme.charts[id]
      data.backgroundColor = self._report.theme.charts[id]
      data.fillOpacity = opacity
    return data

  def add_dataset(self, data, label, colors=None, opacity=0.8):
    """

    """
    data = self.new_dataset(len(self._datasets), data, label, colors, opacity=opacity)
    self._datasets.append(data)
    return data

  @property
  def _js__convertor__(self):
    return '''
      if(data.python){
        result = {datasets: [], labels: data.series};
        data.datasets.forEach(function(rec, i){
          result.datasets.push( {label: data.series[i], data: rec, backgroundColor: options.colors[i]} )
        })
      } else {
        var temp = {}; var labels = [];
        options.y_columns.forEach(function(series){temp[series] = []});
        data.forEach(function(rec){ 
          options.y_columns.forEach(function(name){
            if(rec[name] !== undefined){
              labels.push(rec[options.x_column]); var r = 2; if((options.rDim != undefined) && (rec[options.rDim] != undefined)){r = rec[options.rDim]};
              temp[name].push({y: rec[name], x: rec[options.x_column], r: r})}})});
        result = {datasets: [], labels: labels};
        options.y_columns.forEach(function(series, i){
          dataSet = {label: series, data: [], backgroundColor: options.colors[i]};
          labels.forEach(function(x, i){dataSet.data = temp[series]}); 
        result.datasets.push(dataSet)})
      }'''


class ChartBar(ChartLine):

  def __init__(self, report, width, height, htmlCode, options, profile):
    super(ChartBar, self).__init__(report, width, height, htmlCode, options, profile)
    self._attrs['type'] = 'bar'

  @property
  def options(self):
    """

    :rtype: OptionsPolar
    """
    if self._options is None:
      self._options = OptChartJs.OptionsBar(self._report, attrs=self._options_init)
    return self._options

  def new_dataset(self, id, data, label, colors=None, opacity=0.8, type=None):
    if type is not None:
      data = JsChartJs.DataSetBar(self._report, attrs={"data": data, 'type': self._attrs['type']})
    else:
      data = JsChartJs.DataSetBar(self._report, attrs={"data": data})
    data.label = label
    if colors is None:
      data.backgroundColor = self._report.theme.charts[id]
      data.fillOpacity = opacity
    return data

  def add_dataset(self, data, label, type=None, colors=None, opacity=0.8):
    """

    """
    data = self.new_dataset(len(self._datasets), data, label, colors, opacity=opacity, type=type)
    self._datasets.append(data)
    return data


class ChartPolar(Chart):

  def __init__(self, report, width, height, htmlCode, options, profile):
    super(ChartPolar, self).__init__(report, width, height, htmlCode, options, profile)
    self._attrs['type'] = 'polarArea'

  @property
  def options(self):
    """

    :rtype: OptionsPolar
    """
    if self._options is None:
      self._options = OptChartJs.OptionsPolar(self._report, attrs=self._options_init)
    return self._options

  def new_dataset(self, id, data, label, colors=None, type=None):
    if type is not None:
      data = JsChartJs.DataSetPolar(self._report, attrs={"data": data, 'type': type})
    else:
      data = JsChartJs.DataSetPolar(self._report, attrs={"data": data})
    data.label = label
    data.backgroundColor = self._report.theme.charts
    return data

  def add_dataset(self, data, label, colors=None, opacity=0.4, type=None):
    """

    :return:
    """
    data = self.new_dataset(len(self._datasets), data, label, colors)
    self._datasets.append(data)
    return data

  @property
  def _js__convertor__(self):
    return '''
      if(data.python){
        result = {datasets: [], labels: data.series};
        data.datasets.forEach(function(rec, i){
          var dataset = {label: data.series[i], data: rec, backgroundColor: options.colors, borderColor: options.colors};
          for(var attr in options.attrs){dataset[attr] = options.attrs[attr]};
          result.datasets.push(dataset)
        })
      } else {
        var temp = {}; var labels = []; var uniqLabels = {};
        options.y_columns.forEach(function(series){temp[series] = {}});
        data.forEach(function(rec){ 
          options.y_columns.forEach(function(name){
            if(rec[name] !== undefined){
              if (!(rec[options.x_column] in uniqLabels)){labels.push(rec[options.x_column]); uniqLabels[rec[options.x_column]] = true};
              temp[name][rec[options.x_column]] = rec[name]}})});
        result = {datasets: [], labels: labels};
        options.y_columns.forEach(function(series, i){
          dataSet = {label: series, data: [], backgroundColor: options.colors, borderColor: options.colors};
          for(var attr in options.attrs){dataSet[attr] = options.attrs[attr]};
          labels.forEach(function(x){
            if (temp[series][x] == undefined) {dataSet.data.push(null)} else{dataSet.data.push(temp[series][x])}
          }); result.datasets.push(dataSet)})
      }'''


class ChartHBar(ChartBar):

  def __init__(self, report, width, height, htmlCode, options, profile):
    super(ChartBar, self).__init__(report, width, height, htmlCode, options, profile)
    self._attrs['type'] = 'horizontalBar'


class ChartPie(Chart):

  def __init__(self, report, width, height, htmlCode, options, profile):
    super(ChartPie, self).__init__(report, width, height, htmlCode, options, profile)
    self._attrs['type'] = 'pie'

  @property
  def options(self):
    """

    :rtype: OptionsPie
    """
    if self._options is None:
      self._options = OptChartJs.OptionsPie(self._report, attrs=self._options_init)
    return self._options

  def new_dataset(self, id, data, label="", colors=None, opacity=0.8, type=None):
    data = JsChartJs.DataSetPie(self._report, attrs={"data": data})
    if colors is None:
      data.backgroundColor = self._report.theme.charts
      data.fillOpacity = opacity
    return data

  def add_dataset(self, data, label="", colors=None, opacity=0.8):
    """

    :return:
    """
    data = self.new_dataset(len(self._datasets), data, label, colors=colors, opacity=opacity)
    self._datasets.append(data)
    return data

  @property
  def _js__convertor__(self):
    return ''' 
      if(data.python){
        result = {datasets: [], labels: data.series};
        data.datasets.forEach(function(rec, i){
          result.datasets.push( {label: data.series[i], data: rec, backgroundColor: options.colors} ) })
      } else {
        var temp = {}; var labels = []; var uniqLabels = {};
        options.y_columns.forEach(function(series){temp[series] = {}});
        data.forEach(function(rec){ 
          options.y_columns.forEach(function(name){
            if(rec[name] !== undefined){
              if (!(rec[options.x_column] in uniqLabels)){labels.push(rec[options.x_column]); uniqLabels[rec[options.x_column]] = true};
              temp[name][rec[options.x_column]] = rec[name]}})});
        result = {datasets: [], labels: labels};
        options.y_columns.forEach(function(series){
          dataSet = {label: series, data: [], backgroundColor: []};
          labels.forEach(function(x, i){
            dataSet.backgroundColor.push(options.colors[i]);
            if(temp[series][x] == undefined) {dataSet.data.push(null)} else{dataSet.data.push(temp[series][x])}
          }); result.datasets.push(dataSet)})
      }
      '''


class ChartRadar(Chart):

  def __init__(self, report, width, height, htmlCode, options, profile):
    super(ChartRadar, self).__init__(report, width, height, htmlCode, options, profile)
    self._attrs['type'] = 'radar'

  def new_dataset(self, id, data, label, colors=None, opacity=0.8, type=None):
    data = JsChartJs.DataSetRadar(self._report, attrs={"data": data})
    data.label = label
    if colors is None:
      data.backgroundColor = self._report.theme.charts[id]
      data.borderColor = self._report.theme.charts[id]
      data.borderWidth = 0.2
      data.fillOpacity = opacity
    return data

  def add_dataset(self, data, label, colors=None, opacity=0.4):
    """

    """
    data = self.new_dataset(len(self._datasets), data, label, colors, opacity)
    self._datasets.append(data)
    return data

  @property
  def _js__convertor__(self):
    return '''
      if(data.python){
        result = {datasets: [], labels: data.series};
        data.datasets.forEach(function(rec, i){
          var dataset = {label: data.series[i], data: rec, backgroundColor: options.colors, borderColor: options.colors[i]};
          for(var attr in options.attrs){dataset[attr] = options.attrs[attr]};
          result.datasets.push(dataset)
        })
      } else {
        var temp = {}; var labels = []; var uniqLabels = {};
        options.y_columns.forEach(function(series){temp[series] = {}});
        data.forEach(function(rec){ 
          options.y_columns.forEach(function(name){
            if(rec[name] !== undefined){
              if (!(rec[options.x_column] in uniqLabels)){labels.push(rec[options.x_column]); uniqLabels[rec[options.x_column]] = true};
              temp[name][rec[options.x_column]] = rec[name]}})});
        result = {datasets: [], labels: labels};
        options.y_columns.forEach(function(series, i){
          dataSet = {label: series, data: [], backgroundColor: options.colors, borderColor: options.colors[i]};
          for(var attr in options.attrs){dataSet[attr] = options.attrs[attr]};
          labels.forEach(function(x){
            if (temp[series][x] == undefined) {dataSet.data.push(null)} else{dataSet.data.push(temp[series][x])}
          }); result.datasets.push(dataSet)})
      }'''


class ChartScatter(Chart):

  def __init__(self, report, width, height, htmlCode, options, profile):
    super(ChartScatter, self).__init__(report, width, height, htmlCode, options, profile)
    self._attrs['type'] = 'scatter'

  def new_dataset(self, id, data, label, colors=None, type=None):
    data = JsChartJs.DataSetScatterLine(self._report, attrs={"data": data})
    data.fill = False
    data.label = label
    if colors is None:
      data.backgroundColor = self._report.theme.charts[id]
      data.borderColor = self._report.theme.charts[id]
    return data

  def add_dataset(self, data, label, colors=None):
    """

    :return:
    """
    data = self.new_dataset(len(self._datasets), data, label, colors)
    self._datasets.append(data)
    return data

  @property
  def _js__convertor__(self):
    return ''' 
      if(data.python){
        result = {datasets: [], labels: data.series};
        data.datasets.forEach(function(rec, i){
          result.datasets.push( {label: data.series[i], data: rec, backgroundColor: options.colors[i]} )
        })
      } else {
        var temp = {}; var labels = [];
        options.y_columns.forEach(function(series){temp[series] = []});
        data.forEach(function(rec){ 
          options.y_columns.forEach(function(name){
            if(rec[name] !== undefined){
              labels.push(rec[options.x_column]); var r = 2; if((options.rDim != undefined) && (rec[options.rDim] != undefined)){r = rec[options.rDim]};
              temp[name].push({y: rec[name], x: rec[options.x_column], r: r})}})});
        result = {datasets: [], labels: labels};
        options.y_columns.forEach(function(series, i){
          dataSet = {label: series, data: [], backgroundColor: options.colors[i]};
          labels.forEach(function(x, i){dataSet.data = temp[series]}); 
        result.datasets.push(dataSet)})
      
      }
    '''


class ChartExts(ChartPie):

  def __init__(self, report, width, height, htmlCode, options, profile):
    super(ChartExts, self).__init__(report, width, height, htmlCode, options, profile)
    self.jsImports.add(options['npm'])
    self._attrs['type'] = options['type']




"""

"""

from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsString
from epyk.core.js.primitives import JsObject
from epyk.core.js.primitives import JsBoolean


class JsGeolocation(object):

  def __init__(self, rptObj):
    self._rptObj = rptObj

  def getCurrentPosition(self, callbackFnc, errorFnc=None, options=None):
    """
    The getCurrentPosition() method is used to return the user's position

    Related Pages:

			https://developer.mozilla.org/fr/docs/Web/API/Geolocation/getCurrentPosition
    https://www.w3schools.com/html/html5_geolocation.asp
    https://developer.kaiostech.com/api/geolocation/getposition

    :param callbackFnc: A callback function that takes a Position object as its sole input parameter.
    :param errorFnc: Optional An optional callback function that takes a PositionError object as its sole input parameter.
    :param options: Optional An optional PositionOptions object.
    """
    if errorFnc is None:
      return JsObject.JsObject.new("navigator.geolocation.getCurrentPosition(%s)" % callbackFnc, isPyData=False)

    return JsObject.JsObject.new("navigator.geolocation.getCurrentPosition(%s, %s)" % (callbackFnc, errorFnc), isPyData=False)

  def watchPosition(self, callbackFnc, watchId, errorFnc=None, options=None):
    """
    Returns the current position of the user and continues to return updated position as the user moves

    Example
    options = {"enableHighAccuracy": False, "timeout": 5000, "maximumAge": 0}

    Related Pages:

			https://www.w3schools.com/html/html5_geolocation.asp

    :param callbackFnc: A callback function that takes a Position object as an input parameter.
    :param watchId: The ID number returned by the Geolocation.watchPosition() method when installing the handler you wish to remove.
    :param errorFnc: Optional An optional callback function that takes a PositionError object as an input parameter.
    :param options: Optional An optional PositionOptions object that provides configuration options for the location watch.
    """
    return JsObject.JsObject.new("navigator.geolocation.watchPosition(%s)" % callbackFnc, isPyData=False).setVar(watchId)

  def clearWatch(self, watchId):
    """
    Stops the watchPosition() method

    Related Pages:

			https://www.w3schools.com/html/html5_geolocation.asp

    :param watchId: The ID number returned by the Geolocation.watchPosition() method when installing the handler you wish to remove.
    """
    watchId = JsUtils.jsConvertData(watchId, None)
    return JsObject.JsObject("navigator.geolocation.clearWatch(%s)" % watchId, isPyData=False)


class JsNavigator(object):

  def __init__(self, rptObj):
    self._rptObj = rptObj

  @property
  def geolocation(self):
    """
    The HTML Geolocation API is used to locate a user's position.

    Related Pages:

			https://w3c.github.io/geolocation-api/#navi-geo
    https://www.w3schools.com/html/html5_geolocation.asp
    """
    return JsGeolocation(self._rptObj)

  @property
  def language(self):
    """
    The language property returns the language version of the browser.

    Related Pages:

			https://www.w3schools.com/jsref/prop_nav_language.asp

    :return:
    """
    return JsString.JsString("navigator.language", isPyData=False)

  @property
  def browserLanguage(self):
    """
    The language property returns the language version of the browser.
    For IE10 and earlier versions, you can use the browserLanguage property

    Related Pages:

			https://www.w3schools.com/jsref/prop_nav_language.asp

    :return:
    """
    return JsString.JsString("navigator.browserLanguage", isPyData=False)

  @property
  def appCodeName(self):
    """
    The appCodeName property returns the application code name of the browser

    Related Pages:

			https://www.w3schools.com/js/js_window_navigator.asp
    """
    return JsString.JsString("navigator.appCodeName", isPyData=False)

  @property
  def appName(self):
    """
    The appName property returns the application name of the browser

    Related Pages:

			https://www.w3schools.com/js/js_window_navigator.asp
    """
    return JsString.JsString("navigator.appName", isPyData=False)

  @property
  def product(self):
    """
    The product property returns the product name of the browser engine

    Related Pages:

			https://www.w3schools.com/js/js_window_navigator.asp
    """
    return JsString.JsString("navigator.product", isPyData=False)

  @property
  def appVersion(self):
    """
    The appVersion property returns version information about the browser

    Related Pages:

			https://www.w3schools.com/js/js_window_navigator.asp
    """
    return JsString.JsString("navigator.appVersion", isPyData=False)

  @property
  def cookieEnabled(self):
    """
    """
    return JsString.JsString("navigator.cookieEnabled", isPyData=False)

  @property
  def onLine(self):
    """
    The onLine property returns true if the browser is online

    Related Pages:

			https://www.w3schools.com/js/js_window_navigator.asp
    """
    return JsString.JsString("navigator.onLine", isPyData=False)

  @property
  def platform(self):
    """
    The platform property returns the browser platform (operating system)

    Related Pages:

			https://www.w3schools.com/js/js_window_navigator.asp
    """
    return JsString.JsString("navigator.platform", isPyData=False)

  @property
  def userAgent(self):
    """
    The userAgent property returns the user-agent header sent by the browser to the server

    Related Pages:

			https://www.w3schools.com/js/js_window_navigator.asp

    """
    return JsString.JsString("navigator.userAgent", isPyData=False)

  def javaEnabled(self):
    """
    The javaEnabled() method returns a Boolean value that specifies whether the browser has Java enabled.

    Related Pages:

			https://www.w3schools.com/jsref/met_nav_javaenabled.asp

    :return:
    """
    return JsBoolean.JsBoolean("navigator.javaEnabled()", isPyData=False)



"""
CSS Style module for the HR components
"""

from epyk.core.css.styles.classes import CssStyle


class CssHr(CssStyle.Style):
  _attrs = {'display': 'block', 'border-style': 'inset', 'border-width': '1px', 'margin': '5px'}
  classname = False
  _selector = 'hr'




from epyk.core.js.packages import packageImport
from epyk.core.data.DataClass import DataGroup


class ExtsMutators(DataGroup):

  @packageImport('tabulator-mutators-inputs')
  def number(self, green=None, red=None, threshold=0, cssMapping=None, **kwargs):
    """
    Description:
    -----------

    :param green:
    :param red:
    :param css:
    :param kwargs:
    """
    self._attrs["mutator"] = 'formatNumbers'
    if cssMapping is None:
      self._attrs["mutatorParams"] = {True: {}, False: {}}
    else:
      self._attrs["mutatorParams"] = cssMapping
    self._attrs["mutatorParams"]['threshold'] = threshold
    self._attrs["mutatorParams"][False]['color'] = red or self._report.theme.danger[0]
    self._attrs["mutatorParams"][True]['color'] = green or self._report.theme.success[0]
    if kwargs:
      self._attrs["mutatorParams"].update(kwargs)
    return self

  @packageImport('tabulator-mutators-inputs')
  def text(self, cssMapping, **kwargs):
    """
    Description:
    -----------

    :param green:
    :param red:
    :param css:
    :param kwargs:
    """
    self._attrs["mutator"] = 'formatStrings'
    self._attrs["mutatorParams"] = {'cssMapping': cssMapping}
    if kwargs:
      self._attrs["mutatorParams"].update(kwargs)
    return self

  def custom(self, mutator, mutatorParams, moduleAlias):
    """
    Description:
    -----------

    http://tabulator.info/docs/4.0/mutators

    :param mutator:
    :param mutatorParams:
    :param moduleAlias:
    """
    self._report.jsImports.add(moduleAlias)
    self._attrs["mutator"] = mutator
    self._attrs['mutatorParams'] = mutatorParams
    return self




from epyk.core.js import Js
from epyk.core.js import JsUtils


# ------------------------------------------------------------------------------------------------------------------
#
f = JsUtils.JsFile("jsWindows", path=r"../outs")
jsObj = Js.JsBase()
f.writeJs([
  jsObj.window.btoa("Test").setVar("bin"),
  jsObj.createElement("div", "MyDiv").text(jsObj.objects.get("bin")).css({"color": 'red'}),
  jsObj.body.appendChild(jsObj.objects.get("MyDiv")),
  jsObj.console.log(jsObj.objects.get("bin")),
  jsObj.console.log(jsObj.window.atob(jsObj.objects.get("bin"))),
  jsObj.window.setInterval([jsObj.console.log(jsObj.math.random())], 500).setVar("interva1"),
  jsObj.window.clearInterval(jsObj.objects.get("interva1")),
  jsObj.sessionStorage.setItem("lastname", jsObj.objects.get("bin")),
  jsObj.sessionStorage.removeItem("lastname"),
  jsObj.console.log(jsObj.sessionStorage["lastname"])
])
f.close()


# ------------------------------------------------------------------------------------------------------------------
#
jsObj = Js.JsBase()
f = JsUtils.JsFile("Maths", path=r"../outs")
f.writeJs([
  jsObj.objects.number.new(23.6, varName="MyNumber"),
  jsObj.console.log(jsObj.math.log(2)),
  jsObj.console.log(jsObj.math.E),
  jsObj.console.log(jsObj.math.LN2),
  jsObj.console.log(jsObj.math.LN10),
  jsObj.console.log(jsObj.math.SQRT1_2),
  jsObj.console.log(jsObj.math.random()),
  jsObj.console.log(jsObj.math.random(10, 100)),
  jsObj.console.log(jsObj.math.min(10, 45, 100, -3, 56)),
  jsObj.console.log(jsObj.math.max(10, 45, 100, -3, 56)),
  jsObj.console.log(jsObj.math.floor(jsObj.objects.number.get("MyNumber"))),
  jsObj.console.log(jsObj.math.trunc(jsObj.objects.number.get("MyNumber"))),
  jsObj.console.log(jsObj.math.round(jsObj.objects.number.get("MyNumber"))),
  jsObj.console.log(jsObj.math.sqrt(jsObj.objects.number.get("MyNumber"))),
  jsObj.console.log(jsObj.math.ceil(jsObj.objects.number.get("MyNumber"))),
  jsObj.console.log(jsObj.math.pow(jsObj.objects.number.get("MyNumber"), 2)),
])
f.close()


# ------------------------------------------------------------------------------------------------------------------
#
f = JsUtils.JsFile("Peformance", path=r"../outs")
jsObj = Js.JsBase()
f.writeJs([
  jsObj.performance.mark("youpi"),
  jsObj.performance.measure("youpi"),
  jsObj.console.log(jsObj.performance.now),
  jsObj.console.log(jsObj.performance.getEntriesByName("youpi")),
  jsObj.console.log(jsObj.performance.toJSON()),
  jsObj.console.log(jsObj.performance.now),
])

f.close()


# ------------------------------------------------------------------------------------------------------------------
#
f = JsUtils.JsFile('Location', path=r"../outs")
jsObj = Js.JsBase()
f.writeJs([
  jsObj.console.log(jsObj.location.hostname),
  jsObj.console.log(jsObj.location.pathname),
  jsObj.console.log(jsObj.location.hash),
])
f.close()


# ------------------------------------------------------------------------------------------------------------------
#
jsObj = Js.JsBase()
f = JsUtils.JsFile('Numbers', path=r"../outs")
f.writeJs([
  jsObj.objects.NaN,
  jsObj.console.log(jsObj.objects.NaN),
  jsObj.console.log(jsObj.objects.number.POSITIVE_INFINITY()),
  jsObj.console.log(jsObj.objects.number.MAX_VALUE()),

])
f.close()



from epyk.core.css import Defaults_css
from epyk.core.css.styles.attributes import Attrs


class AttrButton(Attrs):
  def __init__(self, rptObj):
    super(AttrButton, self).__init__(rptObj)
    self.font_size = Defaults_css.font()


class AttrBadge(Attrs):
  def __init__(self, rptObj):
    super(AttrBadge, self).__init__(rptObj)
    self.padding = "1px 3px"
    self.margin = "1px 1px 1px 2px"
    self.vertical_align = "bottom"
    self.font_size = Defaults_css.font(-3)




"""

"""

from epyk.core.js.primitives import JsObject
from epyk.core.js import JsUtils


class JsAttributes(JsObject.JsObject):
  """

  Related Pages:

			https://www.w3schools.com/jsref/dom_obj_attributes.asp

  """

  @property
  def name(self):
    """
    The name property returns the name of the attribute.

    This property is read-only.

    Related Pages:

			https://www.w3schools.com/jsref/prop_attr_name.asp

    :return: A String, representing the name of the attribute
    """
    return "name"

  @property
  def value(self):
    """
    The value property sets or returns the value of the attribute.

    Related Pages:

			https://www.w3schools.com/jsref/prop_attr_value.asp

    :return: Specifies the value of the attribute
    """
    return "value"

  @property
  def specified(self):
    """
    The specified property returns true if the attribute is specified.

    Related Pages:

			https://www.w3schools.com/jsref/prop_attr_specified.asp

    :return: A Boolean, returns true if the attribute is specified, otherwise false
    """
    return "specified"

  def item(self, i):
    """
    The item() method returns the node at the specified index in a NamedNodeMap, as a Node object.

    Related Pages:

			https://www.w3schools.com/jsref/met_namednodemap_item.asp

    :param i: Required. The index of the node in the NamedNodeMap you want to return

    :return: A Node object, representing the attribute node at the specified index.
    """
    return "item(%s)" % i

  def removeNamedItem(self, type):
    """
    The removeNamedItem() method removes the node with the specified name in a NamedNodeMap object.

    Related Pages:

			https://www.w3schools.com/jsref/met_namednodemap_removenameditem.asp

    :param type: Required. The name of the node in the namedNodeMap you want to remove
    :return: 	A Node object, representing the removed attribute node
    """
    return "removeNamedItem(%s)" % type

  def setNamedItem(selfm, jsAttr):
    """
    The setNamedItem() method adds the specified node to the NamedNodeMap.

    If the node already exists, it will be replaced, and the replaced node will be the return value, otherwise the return value will be null.

    Related Pages:

			https://www.w3schools.com/jsref/met_namednodemap_setnameditem.asp

    :return: A Node object, representing the replaced node (if any), otherwise null
    """
    return "setNamedItem(%s)" % jsAttr

  def getNamedItem(self, nodeName):
    """
    The getNamedItem() method returns the attribute node with the specified name from a NamedNodeMap object.

    Related Pages:

			https://www.w3schools.com/jsref/met_namednodemap_getnameditem.asp

    :param nodeName: Required. The name of the node in the namedNodeMap you want to return

    :return: A Node object, representing the attribute node with the specified name
    """
    return "getNamedItem(%s)" % nodeName





class DataEvents(object):

  @property
  def files(self):
    """

    :return:
    """
    from epyk.core.js.primitives import JsObjects
    return JsObjects.JsArray.JsArray.get("Array.from(event.dataTransfer.files)")

  @property
  def data(self):
    """
    Description:
    ------------
    Interface to a standard data object available in any Event.
    This is the default variable name in all the JavaScript embedded methods
    """
    from epyk.core.js.primitives import JsObjects
    return JsObjects.JsObjects.get("data")

  @property
  def event(self):
    """
    Description:
    ------------
    Interface to the standard event
    """
    from epyk.core.js.objects import JsEvents

    return JsEvents.Event()

  @property
  def mouse(self):
    """
    Description:
    ------------
    Interface to the standard mouse event
    """
    from epyk.core.js.objects import JsEvents

    return JsEvents.MouseEvent()

  @property
  def ui(self):
    """
    Description:
    ------------
    Interface to the UI generic event
    """
    from epyk.core.js.objects import JsEvents

    return JsEvents.UIEvent()

  @property
  def touch(self):
    """
    Description:
    ------------
    Interface to a standard touch event.
    This object is available in any event specific to touch screens
    """
    from epyk.core.js.objects import JsEvents

    return JsEvents.TouchEvent()

  @property
  def key(self):
    """
    Description:
    ------------
    Interface to a standard keyboard event.
    This object is available in any keyup, keydown... events
    """
    from epyk.core.js.objects import JsEvents

    return JsEvents.KeyboardEvent()


class DataFile(object):

  @property
  def name(self):
    """

    :return:
    """
    from epyk.core.js.primitives import JsObjects
    return JsObjects.JsString.JsString.get("value.name")

  @property
  def size(self):
    """

    :return:
    """
    from epyk.core.js.primitives import JsObjects
    return JsObjects.JsString.JsString.get("value.size")

  @property
  def lastModifiedDate(self):
    """

    :return:
    """
    from epyk.core.js.primitives import JsObjects
    return JsObjects.JsString.JsString.get("value.lastModifiedDate")

  @property
  def lastModified(self):
    """

    :return:
    """
    from epyk.core.js.primitives import JsObjects
    return JsObjects.JsString.JsString.get("value.lastModified")


class DataLoops(object):

  @property
  def value(self):
    """
    Description:
    -----------
    The value returned by forEach statement.

    Note. For nested loop make sure you store the important information in new variable names.
    """
    from epyk.core.js.primitives import JsObjects
    return JsObjects.JsObject.JsObject.get("value")

  @property
  def dom(self):
    """
    Description:
    -----------

    :return:
    """
    from epyk.core.js.objects import JsNodeDom
    return JsNodeDom.JsDoms.new(varName="value", setVar=False)

  @property
  def i(self):
    """
    Description:
    ------------
    The index value return in loop statement
    """
    from epyk.core.js.primitives import JsObjects
    return JsObjects.JsNumber.JsNumber.get("index")

  @property
  def file(self):
    return DataFile()


class DataPrimitives(object):

  def list(self, data=None, name=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param data: List. The Python object used to feed the list
    :param name: String. The variable name used on the JavaScript
    """
    from epyk.core.js.primitives import JsObjects

    if data is not None:
      setVar = True if name is not None else False
      return JsObjects.JsArray.JsArray(data, varName=name, setVar=setVar)

    return JsObjects.JsArray.JsArray.get(name)

  def dict(self, data=None, name=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param data:
    :param name: String. The variable name used on the JavaScript
    """
    from epyk.core.js.primitives import JsObjects

    if data is not None:
      setVar = True if name is not None else False
      return JsObjects.JsObject.JsObject(data, varName=name, setVar=setVar)

    return JsObjects.JsObject.JsObject.get(name)

  def str(self, data=None, name=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param data:
    :param name: String. The variable name used on the JavaScript
    """
    from epyk.core.js.primitives import JsObjects

    if data is not None:
      setVar = True if name is not None else False
      return JsObjects.JsString.JsString(data, varName=name, setVar=setVar)

    return JsObjects.JsString.JsString.get(name)

  def float(self, data=None, name=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param data:
    :param name: String. The variable name used on the JavaScript
    """
    from epyk.core.js.primitives import JsObjects

    if data is not None:
      setVar = True if name is not None else False
      return JsObjects.JsNumber.JsNumber(data, varName=name, setVar=setVar)

    return JsObjects.JsNumber.JsNumber.get(name)

  def int(self, data=None, name=None):
    """
    Description:
    -----------

    Attributes:
    ----------
    :param data:
    :param name: String. The variable name used on the JavaScript
    """
    from epyk.core.js.primitives import JsObjects

    if data is not None:
      setVar = True if name is not None else False
      return JsObjects.JsNumber.JsNumber(data, varName=name, setVar=setVar)

    return JsObjects.JsNumber.JsNumber.get(name)




from epyk.core.css.catalogs import Catalog

from epyk.core.css.styles.classes import CssStylesDivComms, CssStylesText


class CatalogText(Catalog.CatalogGroup):
  def bold(self):
    """
    Description:
    ------------
    Basic style for an input component
    """
    return self._set_class(CssStylesText.CssTextBold)

  def formula(self):
    """
    Description:
    ------------
    CSS Style for the formulas component
    """
    return self._set_class(CssStylesText.CssFormula)

  def title_1(self):
    """
    Description:
    ------------

    """
    return self._set_class(CssStylesText.CssTitle1)

  def title_2(self):
    """
    Description:
    ------------

    """
    return self._set_class(CssStylesText.CssTitle2)

  def title_3(self):
    """
    Description:
    ------------

    """
    return self._set_class(CssStylesText.CssTitle3)

  def title_4(self):
    """
    Description:
    ------------
    """
    return self._set_class(CssStylesText.CssTitle4)

  def title(self):
    """
    Description:
    ------------

    """
    return self._set_class(CssStylesText.CssTitle)

  def number(self):
    """
    Description:
    ------------

    """
    return self._set_class(CssStylesText.CssNumberCenter)

  def red(self):
    """
    Description:
    ------------

    """
    return self._set_class(CssStylesText.CssMarkRed)

  def colored(self):
    """
    Description:
    ------------
    This class will change the font color to the 7th color of the colors theme.
    The font weight will be set to bold also.
    """
    return self._set_class(CssStylesText.CssMarkBlue)

  def with_border(self):
    """
    Description:
    ------------
    """
    return self._set_class(CssStylesText.CssTextWithBorder)

  def check_mark(self):
    """
    Description:
    ------------
    """
    return self._set_class(CssStylesText.CssCheckMark)

  def item(self):
    """
    Description:
    ------------
    """
    return self._set_class(CssStylesText.CssTextItem)

  def no_selection(self):
    """
    Description:
    ------------
    Text cannot be selected
    """
    return self._set_class(CssStylesText.CssTextNotSelectable)

  def content_editable(self):
    """
    Description:
    ------------
    """
    return self._set_class(CssStylesDivComms.CssContentEditable)
  

class CatalogComments(Catalog.CatalogGroup):
  def header(self):
    """
    Description:
    ------------
    """
    return self._set_class(CssStylesDivComms.CssCommHeader)

  def input(self):
    """
    Description:
    ------------
    """
    return self._set_class(CssStylesDivComms.CssCommInput)

  def content_editable(self):
    """
    Description:
    ------------
    """
    return self._set_class(CssStylesDivComms.CssContentEditable)



"""
Related Pages:

		https://www.w3schools.com/charsets/ref_utf_currency.asp
"""

EURO_CURRENCY_SIGN = '&#8352'
COLON_SIGN = '&#8353'
CRUZEIRO_SIGN = '&#8354'
FRENCH_FRANC_SIGN = '&#8355'
LIRA_SIGN = '&#8356'
MILL_SIGN = '&#8357'
NAIRA_SIGN = '&#8358'
PESETA_SIGN = '&#8359'
RUPEE_SIGN = '&#8360'
WON_SIGN = '&#8361'
NEW_SHEQEL_SIGN = '&#8362'
DONG_SIGN = '&#8363'
EURO_SIGN = '&#8364'
KIP_SIGN = '&#8365'
TUGRIK_SIGN = '&#8366'
DRACHMA_SIGN = '&#8367'
GERMAN_PENNY_SYMBOL = '&#8368'
PESO_SIGN = '&#8369'
GUARANI_SIGN = '&#8370'
AUSTRAL_SIGN = '&#8371'
HRYVNIA_SIGN = '&#8372'
CEDI_SIGN = '&#8373'
LIVRE_TOURNOIS_SIGN = '&#8374'
SPESMILO_SIGN = '&#8375'
TENGE_SIGN = '&#8376'
INDIAN_RUPEE_SIGN = '&#8377'
TURKISH_LIRA_SIGN = '&#8378'
NORDIC_MARK_SIGN = '&#8379'
MANAT_SIGN = '&#8380'
RUBLE_SIGN = '&#8381'
LARI_SIGN = '&#8382'
BITCOIN_SIGN = '&#8383'





from epyk.core.html.options import Options


class OptionsSparkLine(Options):

  @property
  def animate(self):
    """

    """
    return self._config_get(False)

  @animate.setter
  def animate(self, value):
    self._config(value)




import hashlib

from epyk.core.html import Html

from epyk.core.js.packages import JsTabulator
from epyk.core.js.primitives import JsObjects

from epyk.core.data.DataClass import DataClass
from epyk.core.data.DataClass import DataEnum
from epyk.core.data.DataClass import DataGroup
from epyk.core.data.DataClass import DataEnumMulti

# The list of CSS classes
from epyk.core.css.styles import GrpClsTable


class Table(Html.Html):
  requirements = ('tabulator', )
  name = 'Tabulator Table'

  def __init__(self, report, records, width, height, htmlCode, options, profile):
    data, columns, self.__config = [], [], None
    super(Table, self).__init__(report, [], htmlCode=htmlCode, css_attrs={"width": width, "height": height}, profile=profile)
    if records is not None:
      self.config.data = records
    self.style.css.background = None

  @property
  def style(self):
    if self._styleObj is None:
      self._styleObj = GrpClsTable.Tabulator(self)
    return self._styleObj

  @property
  def tableId(self):
    """
    Return the Javascript variable of the chart
    """
    return "%s_obj" % self.htmlCode

  @property
  def config(self):
    if self.__config is None:
      self.__config = TableConfig(self._report)
    return self.__config

  @property
  def js(self):
    """
    Return the Javascript internal object

    :return: A Javascript object

    :rtype: JsTabulator.Tabulator
    """
    if self._js is None:
      self._js = JsTabulator.Tabulator(self._report, selector=self.tableId, setVar=False, parent=self)
    return self._js

  def add_column(self, field, title=None):
    """

    :param field:
    :param title:
    """
    col_def = self.config.columns
    col_def.field = field
    col_def.title = field if title is None else title
    return col_def

  def get_column(self, by_title):
    """

    :param by_title:

    :rtype: Column
    """
    for c in self.config._attrs.get('columns', []):
      if c.title == by_title:
        return c

    return None

  def build(self, data=None, options=None, profile=False):
    if data:
      return self.js.setData(data)

    return 'var %s =  new Tabulator("#%s", %s)' % (self.tableId, self.htmlCode, self.config)

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return "<div %s></div>" % (self.get_attrs(pyClassNames=self.style.get_classes()))


class EnumLayout(DataEnum):

  def fitDataStretch(self):
    """
    Description:
    -----------
    The fitDataStretch layout mode functions in the same way as the fitDataFill mode, but instead of stretching the empty row to fill the table it stretches the last visible column.

    Related Pages:
http://tabulator.info/docs/4.5/layout
    """
    return self.set()

  def fitColumns(self):
    """
    Description:
    -----------
    As an alternative to the default data fit, you can use the fitColumns layout mode to cause Tabulator to resize columns so they fit perfectly in the available table width.

    Related Pages:
http://tabulator.info/docs/4.5/layout
    """
    return self.set()

  def fitDataStretch(self):
    """
    Description:
    -----------
    The fitDataStretch layout mode functions in the same way as the fitDataFill mode, but instead of stretching the empty row to fill the table it stretches the last visible column.

    Related Pages:
http://tabulator.info/docs/4.5/layout
    """
    return self.set()

  def fitDataFill(self):
    """
    Description:
    -----------
    The fitDataFill layout mode functions in the same way as the fitData mode, but ensures that rows are always at least the full width of the table.

    Related Pages:
http://tabulator.info/docs/4.5/layout
    """
    return self.set()


class EnumSorter(DataEnum):

  def string(self):
    """
    Description:
    -----------
    Sorts column as strings of characters

    Related Pages:
http://tabulator.info/examples/4.5#sorters
    """
    return self.set("string")

  def number(self):
    """
    Description:
    -----------
    Sorts column as numbers (integer or float, will also handle numbers using "," separators)

    Related Pages:
http://tabulator.info/examples/4.5#sorters
    """
    return self.set("number")

  def alphanum(self):
    """
    Description:
    -----------
    Sorts column as alpha numeric code

    Related Pages:
http://tabulator.info/examples/4.5#sorters
    """
    return self.set("alphanum")

  def boolean(self):
    """
    Description:
    -----------
    Sorts column as booleans

    Related Pages:
http://tabulator.info/examples/4.5#sorters
    """
    return self.set("boolean")

  def date(self):
    """
    Description:
    -----------
    Sorts column as dates

    Related Pages:
http://tabulator.info/examples/4.5#sorters
    """
    return self.set("date")

  def time(self):
    """
    Description:
    -----------
    Sorts column as dates

    Related Pages:
sorts column as times
    """
    return self.set("time")


class EnumColCss(DataEnumMulti):
  js_conversion = True
  delimiter = " "

  def center(self):
    """
    Description:
    -----------
    CSS Class to center the results
    """
    self._report.body.style.custom_class({'_attrs': {'text-align': 'center'}}, classname="tb-center")
    return self.set("tb-center")

  def color(self, color):
    """
    Description:
    -----------
    CSS Class to change the font color

    Attributes:
    ----------
    :param color: String. The CSS Color
    """
    self._report.body.style.custom_class({'_attrs': {'color': color}}, classname="tb-color-%s" % color)
    return self.set("tb-color-%s" % color)

  def background(self, color):
    """
    Description:
    -----------
    CSS Class to change the background color

    Attributes:
    ----------
    :param color: String. The CSS Color
    """
    self._report.body.style.custom_class({'_attrs': {'background-color': color}}, classname="tb-background-%s" % color)
    return self.set("tb-background-%s" % color)

  def css(self, css_attrs, css_attrs_hover=None):
    """
    Description:
    -----------
    CSS class for bespoke style

    col_def.cssClass.css({'background': 'orange'}, {'background': 'white', 'color': 'blue'})

    Attributes:
    ----------
    :param css_attrs: Dictionary. The CSS attributes for the class
    :param css_attrs_hover: Dictionary. The CSS Hover attributes for the class
    """
    has_style = str(hashlib.sha1(str(css_attrs).encode()).hexdigest())
    self._report.body.style.custom_class({'_attrs': css_attrs, '_hover': css_attrs_hover}, classname="tb-style-%s" % has_style)
    return self.set("tb-style-%s" % has_style)


class PersistenceGroup(DataClass):

  @property
  def groupBy(self):
    """
    persist only the groupBy setting

    :return:
    """
    return self._attrs["groupBy"]

  @groupBy.setter
  def groupBy(self, val):
    self._attrs["groupBy"] = val

  @property
  def groupStartOpen(self):
    return self._attrs["groupStartOpen"]

  @groupStartOpen.setter
  def groupStartOpen(self, val):
    self._attrs["groupStartOpen"] = val

  @property
  def groupHeader(self):
    return self._attrs["groupHeader"]

  @groupHeader.setter
  def groupHeader(self, val):
    self._attrs["groupHeader"] = val


class PersistencePage(DataClass):

  @property
  def size(self):
    """
    Description:
    -----------
    persist the current page size

    Related Pages:
http://tabulator.info/docs/4.5/release#persistence
    """
    return self._attrs["size"]

  @size.setter
  def size(self, val):
    self._attrs["size"] = val

  @property
  def page(self):
    """
    Description:
    -----------
    do not persist the current page

    Related Pages:
http://tabulator.info/docs/4.5/release#persistence
    """
    return self._attrs["page"]

  @page.setter
  def page(self, val):
    self._attrs["page"] = val


class Persistence(DataClass):

  @property
  def sort(self):
    """
    Description:
    -----------
    You can ensure the data sorting is stored for the next page load by setting the sort property of the persistence option to true

    Related Pages:
http://tabulator.info/docs/4.5/release#persistence
    """
    return self._attrs["sort"]

  @sort.setter
  def sort(self, val):
    self._attrs["sort"] = val

  @property
  def filter(self):
    """
    Description:
    -----------
    You can ensure the data filtering is stored for the next page load by setting the filter property of the persistence option to true

    Related Pages:
http://tabulator.info/docs/4.5/release#persistence
    """
    return self._attrs["filter"]

  @filter.setter
  def filter(self, val):
    self._attrs["filter"] = val

  @property
  def group(self):
    """
    Description:
    -----------
    You can ensure the row grouping settings are stored for the next page load by setting the group property of the persistence option to true

    Related Pages:
http://tabulator.info/docs/4.5/release#persistence

    :rtype: PersistenceGroup
    """
    return self.has_attribute(PersistenceGroup)

  @property
  def page(self):
    """
    Description:
    -----------
    You can ensure the pagination settings are stored for the next page load by setting the page property of the persistence option to true

    Related Pages:
http://tabulator.info/docs/4.5/release#persistence

    :rtype: PersistencePage
    """
    return self.has_attribute(PersistencePage)

  @property
  def columns(self):
    """
    Description:
    -----------
    You can ensure the layout of columns is stored for the next page load by setting the columns property of the persistence option to true

    Related Pages:
http://tabulator.info/docs/4.5/release#persistence
    """
    return self._attrs["columns"]

  @columns.setter
  def columns(self, val):
    self._attrs["columns"] = val


class BottomCalcParams(DataClass):

  @property
  def precision(self):
    """
    the number of decimals to display (default is 2), setting this value to false will display however many decimals are provided with the number

    http://tabulator.info/docs/4.5/column-calcs#func-builtin
    """
    return self._attrs["precision"]

  @precision.setter
  def precision(self, val):
    self._attrs["precision"] = val


class ColumnsGroup(DataClass):

  @property
  def title(self):
    return self._attrs["title"]

  @title.setter
  def title(self, val):
    self._attrs["title"] = val

  @property
  def columns(self):
    return self.sub_data_enum("columns", Column)


class Editor(DataGroup):

  def input(self, search=True, elementAttributes=None, **kwargs):
    """
    Description:
    -----------
    The input editor allows entering of a single line of plain text

    Related Pages:
http://tabulator.info/docs/4.5/edit#edit-builtin

    Attributes:
    ----------
    :param search: use search type input element with clear button
    :param elementAttributes: set attributes directly on the input element
    """
    self._attrs["editor"] = 'input'
    self._attrs["editorParams"] = {'search': search}
    if elementAttributes is not None:
      self._attrs["editorParams"][elementAttributes] = elementAttributes
    if kwargs:
      self._attrs["editorParams"].update(kwargs)
    return self

  def textarea(self, verticalNavigation="editor", elementAttributes=None, **kwargs):
    """
    Description:
    -----------
    The textarea editor allows entering of multiple lines of plain text

    Related Pages:
http://tabulator.info/docs/4.5/edit#edit-builtin

    Attributes:
    ----------
    :param verticalNavigation: set attributes directly on the textarea element
    :param elementAttributes: determine how use of the up/down arrow keys will affect the editor,
    """
    self._attrs["editor"] = 'textarea'
    self._attrs["editorParams"] = {'verticalNavigation': verticalNavigation, "elementAttributes": elementAttributes}
    self._attrs["editorParams"].update(self._attrs["editorParams"].pop('kwargs'))
    return self

  def number(self, min=None, max=None, step=1, elementAttributes=None, verticalNavigation="table", **kwargs):
    """
    Description:
    -----------
    The number editor allows for numeric entry with a number type input element with increment and decrement buttons.

    Related Pages:
http://tabulator.info/docs/4.5/edit#edit-builtin

    Attributes:
    ----------
    :param min: the maximum allowed value
    :param max: the minimum allowed value
    :param step: the step size when incrementing/decrementingthe value (default 1)
    :param elementAttributes: set attributes directly on the input element
    :param verticalNavigation: determine how use of the up/down arrow keys will affect the editor,
    :param kwargs:
    """
    self._attrs["editor"] = 'number'
    self._attrs["editorParams"] = {'step': step, 'verticalNavigation': verticalNavigation, "elementAttributes": elementAttributes}
    if min is not None:
      self._attrs["editorParams"]['min'] = min
    if max is not None:
      self._attrs["editorParams"]['max'] = max
    self._attrs["editorParams"].update(self._attrs["editorParams"].pop('kwargs'))
    return self

  def range(self, min=None, max=None, step=1, elementAttributes=None, **kwargs):
    """
    Description:
    -----------
    The range editor allows for numeric entry with a range type input element.

    Related Pages:
http://tabulator.info/docs/4.5/edit#edit-builtin

    Attributes:
    ----------
    :param min: the maximum allowed value
    :param max: the minimum allowed value
    :param step: the step size when incrementing/decrementingthe value (default 1)
    :param elementAttributes: set attributes directly on the input element
    :param kwargs:
    """
    self._attrs["editor"] = 'range'
    self._attrs["editorParams"] = {'step': step, "elementAttributes": elementAttributes}
    if min is not None:
      self._attrs["editorParams"]['min'] = min
    if max is not None:
      self._attrs["editorParams"]['max'] = max
    self._attrs["editorParams"].update(self._attrs["editorParams"].pop('kwargs'))
    return self

  def tick(self, tristate=False, indeterminateValue=None, elementAttributes=None, **kwargs):
    """
    Description:
    -----------
    The tick editor allows for boolean values using a checkbox type input element.

    Related Pages:
http://tabulator.info/docs/4.5/edit#edit-builtin

    Attributes:
    ----------
    :param tristate: allow tristate tickbox (default false)
    :param indeterminateValue:  when using tristate tickbox what value should the third indeterminate state have (default null)
    :param elementAttributes: set attributes directly on the input element
    :param kwargs:
    """
    self._attrs["editor"] = 'tick'
    self._attrs["editorParams"] = {'tristate': tristate, 'indeterminateValue': indeterminateValue}
    if elementAttributes is not None:
      self._attrs["editorParams"]['elementAttributes'] = elementAttributes
    self._attrs["editorParams"].update(self._attrs["editorParams"].pop('kwargs'))
    return self

  def stars(self, elementAttributes=None, **kwargs):
    """
    Description:
    -----------
    The star editor allows entering of numeric value using a star rating indicator.

    This editor will automatically detect the correct number of stars to use if it is used on the same column as the star formatter.

    Related Pages:
http://tabulator.info/docs/4.5/edit#edit-builtin

    Attributes:
    ----------
    :param elementAttributes: set attributes directly on the star holder elemen
    :param kwargs:
    """
    self._attrs["editor"] = 'star'
    self._attrs["editorParams"] = {}
    if elementAttributes is not None:
      self._attrs["editorParams"]['elementAttributes'] = elementAttributes
    self._attrs["editorParams"].update(self._attrs["editorParams"].pop('kwargs'))
    return self

  def select(self, values, listItemFormatter=None, sortValuesList=None, defaultValue=None, elementAttributes=None,
                    verticalNavigation="hybrid", **kwargs):
    """
    Description:
    -----------
    The select editor creates a dropdown select box to allow the user to select from some predefined options passed into the values property of the editorParams option.

    Related Pages:
http://tabulator.info/docs/4.5/edit#edit-builtin

    Attributes:
    ----------
    :param values: a list of values to be displayed to the user
    :param listItemFormatter: a function that should return the HTML contents for each item in the value list
    :param sortValuesList: if values property is set to true this option can be used to set how the generated list should be sorted
    :param defaultValue: set the value that should be selected by default if the cells value is undefined
    :param elementAttributes: set attributes directly on the input element
    :param verticalNavigation: determine how use of the up/down arrow keys will affect the editor,
    :param kwargs:
    """
    self._attrs["editor"] = 'select'
    self._attrs["editorParams"] = {k: v for k, v in locals().items() if k != 'self'}
    self._attrs["editorParams"].update(self._attrs["editorParams"].pop('kwargs'))
    return self

  def autocomplete(self, values, showListOnEmpty=None, freetext=None, allowEmpty=None, searchFunc=None,
                          listItemFormatter=None, sortValuesList=None, defaultValue=None, elementAttributes=None,
                          verticalNavigation=None, **kwargs):
    """
    Description:
    -----------
    The autocomplete editor allows users to search a list of predefined options passed into the values property of the editorParams option.

    Related Pages:
http://tabulator.info/docs/4.5/edit#edit-builtin

    Attributes:
    ----------
    :param values: a list of values to be displayed to the user
    :param showListOnEmpty: show all values in the list when the input element is empty (default false)
    :param freetext: allow the user to press enter to save a value to the cell that is not in the list (default false)
    :param allowEmpty: allow the user to save an empty value to the cell (default false)
    :param searchFunc: unction to search through array of value objects and return those that match the search term
    :param listItemFormatter: a function that should return the HTML contents for each item in the value list
    :param sortValuesList:  if values property is set to true this option can be used to set how the generated list should be sorted
    :param defaultValue: set the value that should be selected by default if the cells value is undefined
    :param elementAttributes: set attributes directly on the input element
    :param verticalNavigation: determine how use of the up/down arrow keys will affect the editor,
    """
    self._attrs["editor"] = 'autocomplete'
    self._attrs["editorParams"] = {k: v for k, v in locals().items() if k != 'self'}
    self._attrs["editorParams"].update(self._attrs["editorParams"].pop('kwargs'))
    return self


class Formattors(DataGroup):

  def text(self, **kwargs):
    """
    Description:
    -----------
    The plaintext formater is the default formatter for all cells and will simply dispay the value of the cell as text.

    Related Pages:
http://tabulator.info/docs/4.1/format

    Attributes:
    ----------
    :param kwargs:
    """
    self._attrs["formatter"] = 'plaintext'
    if kwargs:
      self._attrs["formatterParams"] = kwargs
    return self

  def textarea(self, **kwargs):
    """
    Description:
    -----------
    The textarea formater shows text with carriage returns intact (great for multiline text), this formatter will also adjust the height of rows to fit the cells contents when columns are resized.

    Related Pages:
http://tabulator.info/docs/4.1/format

    Attributes:
    ----------
    :param kwargs:
    """
    self._attrs["formatter"] = 'textarea'
    if kwargs:
      self._attrs["formatterParams"] = kwargs
    return self

  def html(self, **kwargs):
    """
    Description:
    -----------
    he html formater displays un-sanitized html.

    Related Pages:
http://tabulator.info/docs/4.1/format

    Attributes:
    ----------
    :param kwargs:
    """
    self._attrs["formatter"] = 'html'
    if kwargs:
      self._attrs["formatterParams"] = kwargs
    return self

  def money(self, decimal=",", thousand=".", precision=False, symbol=None, symbolAfter=None, **kwargs):
    """
    Description:
    -----------
    The money formater formats a number into currency notation (eg. 1234567.8901 -> 1,234,567.89).

    Related Pages:
http://tabulator.info/docs/4.1/format

    Attributes:
    ----------
    :param decimal: Symbol to represent the decimal point (default ".")
    :param thousand: Symbol to represent the thousands seperator (default ",")
    :param precision: the number of decimals to display (default is 2), setting this value to false will display however many decimals are provided with the number
    :param symbol: currency symbol (no default)
    :param symbolAfter: position the symbol after the number (default false)
    :param kwargs:
    """
    self._attrs["formatter"] = 'formatterParams'
    if kwargs:
      self._attrs["formatterParams"] = kwargs
    return self

  def image(self, height=None, width=None, **kwargs):
    """
    Description:
    -----------
    The image formater creates an img element with the src set as the value. (triggers the normalizeHeight function on the row on image load).

    Related Pages:
http://tabulator.info/docs/4.1/format

    Attributes:
    ----------
    :param height: a CSS value for the height of the image
    :param width: a CSS value for the width of the image
    :param kwargs:
    """
    self._attrs["formatter"] = 'image'
    formatParams = {}
    if height is not None:
      formatParams['height'] = height
    if width is not None:
      formatParams['width'] = width
    self._attrs["formatterParams"] = formatParams
    for k, v in kwargs.items():
      self._attrs["formatterParams"][k] = v
    return self

  def link(self, label=None, url=None, target='_blank', urlPrefix=None, labelField=None, urlField=None, **kwargs):
    """
    Description:
    -----------
    The link formater renders data as an anchor with a link to the given value (by default the value will be used as both the url and the label of the tag).

    Related Pages:
http://tabulator.info/docs/4.1/format

    Attributes:
    ----------
    :param label: a string representing the lable, or a function which must return the string for the label, the function is passed the Cell Component as its first argument
    :param url:  a string representing the url, or a function which must return the string for the url, the function is passed the Cell Component as its first argument
    :param target: a string representing the value of the anchor tags target artibute (eg. set to "_blank" to open link in new tab)
    :param urlPrefix: a prefix to put before the url value (eg. to turn a emaill address into a clickable mailto link you should set this to "mailto:")
    :param labelField: the field in the row data that should be used for the link lable
    :param urlField: the field in the row data that should be used for the link url
    :param kwargs:
    """
    self._attrs["formatterParams"] = {k: v for k, v in locals().items() if k != 'self' and v is not None}
    self._attrs["formatterParams"].update(self._attrs["formatterParams"].pop('kwargs'))
    self._attrs["formatter"] = 'link'
    return self

  def datetime(self, inputFormat="YYYY-MM-DD", outputFormat="YYYY-MM-DD", invalidPlaceholder="(invalid date)", **kwargs):
    """
    Description:
    -----------
    The datetime formater transforms on format of date or time into another.

    Related Pages:
http://tabulator.info/docs/4.1/format

    Attributes:
    ----------
    :param inputFormat:
    :param outputFormat:
    :param invalidPlaceholder:
    :param kwargs:
    """
    self._attrs["formatter"] = 'datetime'
    self._attrs["formatterParams"] = {"inputFormat": inputFormat, "outputFormat": outputFormat, "invalidPlaceholder": invalidPlaceholder}
    if kwargs:
      self._attrs["formatterParams"].update(kwargs)
    return self

  def tickcross(self, allowEmpty=True, allowTruthy=True, tickElement="<i class='fa fa-check'></i>",
                          crossElement="<i class='fa fa-times'></i>", **kwargs):
    """
    Description:
    -----------
    The tickCross formatter displays a green tick if the value is (true|'true'|'True'|1) and a red cross if not.

    Related Pages:
http://tabulator.info/docs/4.1/format

    Attributes:
    ----------
    :param allowEmpty: set to true to cause empty values (undefined, null, "") to display an empty cell instead of a cross (default false)
    :param allowTruthy: set to true to allow any truthy value to show a tick (default false)
    :param tickElement: custom HTML for the tick element, if set to false the tick element will not be shown (it will only show crosses)
    :param crossElement: custom HTML for the cross element, if set to false the cross element will not be shown (it will only show ticks)
    :param kwargs:
    """
    self._attrs["formatter"] = 'tickCross'
    self._attrs["formatterParams"] = {'allowEmpty': allowEmpty, 'allowTruthy': allowTruthy, 'tickElement': tickElement,
                                      'crossElement': crossElement}
    if kwargs:
      self._attrs["formatterParams"].update(kwargs)
    return self

  def color(self, **kwargs):
    """
    Description:
    -----------
    The color formater sets the background colour of the cell to the value. The cell's value can be any valid CSS color eg. #ff0000, #f00, rgb(255,0,0), red, rgba(255,0,0,0), hsl(0, 100%, 50%)

    Related Pages:
http://tabulator.info/docs/4.1/format
    """
    self._attrs["formatter"] = 'color'
    if kwargs:
      self._attrs["formatterParams"] = kwargs
    return self

  def star(self, starts, **kwargs):
    """
    Description:
    -----------
    The star formater displays a graphical star rating based on integer values.

    Related Pages:
http://tabulator.info/docs/4.1/format

    Attributes:
    ----------
    :param starts: maximum number of stars to be displayed (default 5)
    """
    self._attrs["formatter"] = 'star'
    formatParams = {"starts": starts}
    for k, v in kwargs.items():
      formatParams[k] = v
    self._attrs["formatterParams"] = formatParams
    return self

  def progress(self, min=0, max=100, color=None, legend=None, legendColor=None, legendAlign=None, **kwargs):
    """

    :param min:
    :param max:
    :param color:
    :param legend:
    :param legendColor:
    :param legendAlign:
    :param kwargs:
    """
    self._attrs["formatter"] = 'progress'
    if kwargs:
      self._attrs["formatterParams"] = kwargs
    return self

  def lookup(self, data, **kwargs):
    """
    Description:
    -----------
    The lookup formater looks up the value to display from a object passed into the formatterParams property, if not present it displays the current cell value

    Related Pages:
http://tabulator.info/docs/4.1/format

    Attributes:
    ----------
    :param data: Dictionary for the lookup
    """
    self._attrs["formatter"] = 'lookup'
    self._attrs['formatterParams'] = data
    if kwargs:
      self._attrs["formatterParams"].update(kwargs)
    return self

  def custom(self, value):
    """

    :param value:
    """
    self._attrs["formatter"] = JsObjects.JsObjects.get(value)
    return self

  def wrapper(self, formatter, css_attrs, formatterParams=None):
    """

    .formatters.wrapper("progress", {"height": '6px'}, {'color': ['orange', 'green']})
    
    :param formatter:
    :param css_attrs:
    :param formatterParams:
    """
    self._attrs["formatter"] = JsObjects.JsObjects.get('''
      function(cell, formatterParams){ 
        const cssAttrs = formatterParams.css;
        var cell = cell.getTable().modules.format.getFormatter('%s').call(cell.getTable().modules.format, cell, formatterParams);
        let frag = document.createRange().createContextualFragment(cell).firstChild;
        Object.keys(cssAttrs).forEach(function(key){frag.style[key] = cssAttrs[key]}); 
        return frag; }''' % formatter)
    self._attrs['formatterParams'] = formatterParams or {}
    self._attrs['formatterParams']['css'] = css_attrs
    return self


class Validators(DataGroup):

  def required(self):
    """
    Description:
    -----------
    The required validator allows values that are not null or an empty string

    Related Pages:
http://tabulator.info/docs/4.5/validate
    """
    self._attrs["validator"].append('required')
    return self

  def unique(self):
    """
    Description:
    -----------
    The unique validator allows values that do not match the value of any other cell in this column

    Related Pages:
http://tabulator.info/docs/4.5/validate
    """
    self._attrs["validator"].append('unique')
    return self

  def integer(self):
    """
    Description:
    -----------
    The integer validator allows values that are valid integers

    Related Pages:
http://tabulator.info/docs/4.5/validate
    """
    self._attrs["validator"].append('integer')
    return self

  def float(self):
    """
    Description:
    -----------
    The float validator allows values that are valid floats

    Related Pages:
http://tabulator.info/docs/4.5/validate
    """
    self._attrs["validator"].append('float')
    return self

  def numeric(self):
    """
    Description:
    -----------
    The float validator allows values that are valid floats

    Related Pages:
http://tabulator.info/docs/4.5/validate
    """
    self._attrs["validator"].append('numeric')
    return self

  def min(self, val):
    """
    Description:
    -----------
    The min validator allows numeric values that are greater than or equal to parameter

    Related Pages:
http://tabulator.info/docs/4.5/validate
    """
    self._attrs["validator"].append('min:%s' % val)
    return self

  def max(self, val):
    """
    Description:
    -----------
    The max validator allows numeric values that are less than or equal to parameter

    Related Pages:
http://tabulator.info/docs/4.5/validate
    """
    self._attrs["validator"].append('max:%s' % val)
    return self

  def maxLength(self, val):
    """
    Description:
    -----------
    The maxLength validator allows string values that have a length less than or equal to parameter

    Related Pages:
http://tabulator.info/docs/4.5/validate
    """
    self._attrs["validator"].append('maxLength:%s' % val)
    return self

  def list(self, vals):
    """
    Description:
    -----------
    The in validator allows that values that match a value from the | delimieted string in the parameter

    Related Pages:
http://tabulator.info/docs/4.5/validate
    """
    self._attrs["validator"].append('in:%s' % "".join(map(lambda  x: str(), vals)))
    return self

  def regex(self, val):
    """
    Description:
    -----------
    The regex validator allows values that match the supplied regex

    Related Pages:
http://tabulator.info/docs/4.5/validate
    """
    self._attrs["validator"].append('regex:%s' % val)
    return self


class Extensions(DataGroup):

  @property
  def editors(self):
    """
    Description:
    -----------
    The edit module allows the user to change data in cells, header filters are also dependant on this module.

    More information on these functions can be found in the Editing Data Documentation.

    Related Pages:
http://tabulator.info/docs/4.0/modules
    """
    from epyk.core.html.tables.exts import TbEditors

    return TbEditors.ExtsEditors(self._report, self._attrs, parent=self)

  @property
  def formatters(self):
    """
    Description:
    -----------
    You can set cell formatters on a per column basis using the formatter option in the column definition object.

    You can pass an optional additional parameter with the formatter, formatterParams that should contain an object with additional information for configuring the formatter.

    Related Pages:
http://tabulator.info/docs/4.0/format
    """
    from epyk.core.html.tables.exts import TbFormatters

    return TbFormatters.ExtsFormattors(self._report, self._attrs, parent=self)

  @property
  def mutators(self):
    """
    Description:
    -----------
    Mutators are used to alter data as it is parsed into Tabulator.
    For example if you wanted to convert a numeric column into a boolean based on its value, before the data is used to build the table.

    Related Pages:
http://tabulator.info/docs/4.0/mutators
    """
    from epyk.core.html.tables.exts import TbMutators

    return TbMutators.ExtsMutators(self._report, self._attrs, parent=self)

  @property
  def validators(self):
    """
    Description:
    -----------
    The validate module allows for validation of editied data before it is stored in the table.
    More information on these functions can be found in the Validation Documentation.
    This can be extended to add custom validator functions to the default list:

    Related Pages:
http://tabulator.info/docs/4.0/modules
    """
    from epyk.core.html.tables.exts import TbValidators

    return TbValidators.ExtsValidators(self._report, self._attrs, parent=self)


class Column(DataClass):

  @property
  def align(self):
    """
    Description:
    -----------
    sets the text alignment for this column (left|center|right)

    Related Pages:
http://tabulator.info/docs/4.5/columns
    """
    return self._attrs["align"]

  @align.setter
  def align(self, val):
    self._attrs["align"] = val

  @property
  def bottomCalc(self):
    """
    Description:
    -----------
    the column calculation to be displayed at the bottom of this column(see Column Calculations for more details)

    Related Pages:
http://tabulator.info/docs/4.5/columns
    """
    return self._attrs["bottomCalc"]

  @bottomCalc.setter
  def bottomCalc(self, val):
    self._attrs["bottomCalc"] = val

  @property
  def bottomCalcParams(self):
    """
    Description:
    -----------
    additional parameters you can pass to the bottomCalc calculation function(see Column Calculations for more details)

    Related Pages:
http://tabulator.info/docs/4.5/columns

    :rtype: BottomCalcParams
    """
    return self.sub_data("bottomCalcParams", BottomCalcParams)

  @property
  def cssClass(self):
    """

    :rtype: EnumColCss
    """
    return self.has_attribute(EnumColCss)

  @property
  def editable(self):
    """
    Description:
    -----------
    callback to check if the cell is editable (see Manipulating Data for more details)

    Related Pages:
http://tabulator.info/docs/4.5/columns
    """
    return self._attrs["editable"]

  @editable.setter
  def editable(self, val):
    self._attrs["editable"] = val

  @property
  def exts(self):
    """
    Description:
    -----------
    Tabulator is built in a modular fashion with a core codebase providing basic table rendering functionality and a series of modules that provide all of its wonderfull features.

    Related Pages:
http://tabulator.info/docs/4.0/modules
    """
    return Extensions(self._report, self._attrs, parent=self)

  @property
  def editors(self):
    """
    Description:
    -----------
    Tabulator comes with a number of built-in editors including:

    http://tabulator.info/docs/4.5/edit#edit-builtin
    """
    return Editor(self, self._attrs)

  @property
  def field(self):
    """
    Description:
    -----------
    Required (not required in icon/button columns) this is the key for this column in the data array

    Related Pages:
http://tabulator.info/docs/4.5/columns
    """
    return self._attrs["field"]

  @field.setter
  def field(self, val):
    self._attrs["field"] = val

  @property
  def formatters(self):
    """
    Description:
    -----------
    You can set cell formatters on a per column basis using the formatter option in the column definition object.

    Related Pages:
http://tabulator.info/docs/4.5
    """
    return Formattors(self, self._attrs)

  @property
  def frozen(self):
    """
    Description:
    -----------
    freezes the column in place when scrolling (see Frozen Columns for more details)

    Related Pages:
http://tabulator.info/docs/4.5/columns
    """
    return self._attrs["frozen"]

  @frozen.setter
  def frozen(self, val):
    self._attrs["frozen"] = val

  @property
  def headerVertical(self):
    """
    Description:
    -----------
    change the orientation of the column header to vertical (see Vertical Column Headers for more details)

    Related Pages:
http://tabulator.info/docs/4.5/columns
    """
    return self._attrs["headerVertical"]

  @headerVertical.setter
  def headerVertical(self, val):
    self._attrs["headerVertical"] = val

  @property
  def headerSort(self):
    """
    Description:
    -----------
    user can sort by clicking on the header (see Sorting Data for more details)

    Related Pages:
http://tabulator.info/docs/4.5/columns
    """
    return self._attrs["headerSort"]

  @headerSort.setter
  def headerSort(self, val):
    self._attrs["headerSort"] = val

  @property
  def headerVisible(self):
    """
    Description:
    -----------
    By setting the headerVisible option to false you can hide the column headers and present the table as a simple list if needed.

    Related Pages:
http://tabulator.info/docs/4.5/columns
    """
    return self._attrs["headerVisible"]

  @headerVisible.setter
  def headerVisible(self, val):
    self._attrs["headerVisible"] = val

  @property
  def sorter(self):
    """
    Description:
    -----------
    By default Tabulator will attempt to guess which sorter should be applied to a column based on the data contained in the first row.

    Related Pages:
http://tabulator.info/examples/4.5#sorters

    :rtype: EnumSorter
    """
    return self.sub_data("sorter", EnumSorter)

  @property
  def width(self):
    """
    Description:
    -----------
    sets the width of this column, this can be set in pixels or as a percentage of total table width (if not set the system will determine the best)

    Related Pages:
http://tabulator.info/docs/4.5/columns
    """
    return self._attrs["width"]

  @width.setter
  def width(self, val):
    self._attrs["width"] = val

  @property
  def minwidth(self):
    """
    Description:
    -----------
    sets the minimum width of this column, this should be set in pixels (this takes priority over the global option of columnMinWidth)

    Related Pages:
http://tabulator.info/docs/4.1/columns
    """
    return self._attrs["minwidth"]

  @minwidth.setter
  def minwidth(self, val):
    self._attrs["minwidth"] = val

  @property
  def widthGrow(self):
    """
    Description:
    -----------
    when using fitColumns layout mode, determines how much the column should grow to fill available space (see Table Layout for more details)

    Related Pages:
http://tabulator.info/docs/4.5/columns
    """
    return self._attrs["widthGrow"]

  @widthGrow.setter
  def widthGrow(self, val):
    self._attrs["widthGrow"] = val

  @property
  def responsive(self):
    """
    Description:
    -----------
    an integer to determine when the column should be hidden in responsive mode (see Responsive Layout for more details)

    Related Pages:
http://tabulator.info/docs/4.5/columns
    """
    return self._attrs["responsive"]

  @responsive.setter
  def responsive(self, val):
    self._attrs["responsive"] = val

  @property
  def resizable(self):
    """
    Description:
    -----------
    set whether column can be resized by user dragging its edges (see Table Layout for more details)

    Related Pages:
http://tabulator.info/docs/4.5/columns
    """
    return self._attrs["resizable"]

  @resizable.setter
  def resizable(self, val):
    self._attrs["resizable"] = val

  @property
  def title(self):
    """
    Description:
    -----------
    Required This is the title that will be displayed in the header for this column

    Related Pages:
http://tabulator.info/docs/4.5/columns
    """
    return self._attrs["title"]

  @title.setter
  def title(self, val):
    self._attrs["title"] = val

  @property
  def titleFormatter(self):
    """
    Description:
    -----------
    formatter function for header title (see Formatting Data for more details)

    Related Pages:
http://tabulator.info/docs/4.5/columns
    """
    return self._attrs["titleFormatter"]

  @titleFormatter.setter
  def titleFormatter(self, val):
    self._attrs["titleFormatter"] = val

  @property
  def topCalc(self):
    """
    Description:
    -----------
    the column calculation to be displayed at the top of this column(see Column Calculations for more details)

    Related Pages:
http://tabulator.info/docs/4.5/columns
    """
    return self._attrs["topCalc"]

  @topCalc.setter
  def topCalc(self, val):
    self._attrs["topCalc"] = val

  @property
  def validator(self):
    """
    Description:
    -----------
    set the validator to be used to approve data when a user edits a cell. (see Manipulating Data for more details)

    Related Pages:
http://tabulator.info/docs/4.5/validate
    """
    self._attrs["validator"] = []
    return Validators(self, self._attrs)


class TableConfig(DataClass):

  @property
  def ajaxURL(self):
    """
    Description:
    -----------

    Related Pages:
http://tabulator.info/docs/4.0/data
    """
    return self._attrs["ajaxURL"]

  @ajaxURL.setter
  def ajaxURL(self, val):
    self._attrs["ajaxURL"] = val

  @property
  def ajaxProgressiveLoad(self):
    """
    Description:
    -----------

    Related Pages:
http://tabulator.info/docs/4.0/data
    """
    return self._attrs["ajaxProgressiveLoad"]

  @ajaxProgressiveLoad.setter
  def ajaxProgressiveLoad(self, val):
    self._attrs["ajaxProgressiveLoad"] = val

  @property
  def autoColumns(self):
    """
    Description:
    -----------
    If you set the autoColumns option to true, every time data is loaded into the table through the data option or through the setData function, Tabulator will examine the first row of the data and build columns to match that data.

    Related Pages:
http://tabulator.info/docs/4.2/columns
    """
    return self._attrs["autoColumns"]

  @autoColumns.setter
  def autoColumns(self, val):
    self._attrs["autoColumns"] = val

  @property
  def addRowPos(self):
    """
    Description:
    -----------
    The position in the table for new rows to be added, "bottom" or "top"

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["addRowPos"]

  @addRowPos.setter
  def addRowPos(self, val):
    self._attrs["addRowPos"] = val

  @property
  def clipboard(self):
    """
    Description:
    -----------
    Enable clipboard module

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["clipboard"]

  @clipboard.setter
  def clipboard(self, val):
    self._attrs["clipboard"] = val

  @property
  def clipboardPasteAction(self):
    """
    Description:
    -----------
    Clipboard paste action function

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["clipboardPasteAction"]

  @clipboardPasteAction.setter
  def clipboardPasteAction(self, val):
    self._attrs["clipboardPasteAction"] = val

  @property
  def columns(self):
    """
    Description:
    -----------
    Holder for column definition array

    Related Pages:
http://tabulator.info/docs/4.2/options

    :rtype: Column
    """
    return self.sub_data_enum("columns", Column)

  @property
  def columns_group(self):
    """

    Related Pages:
http://tabulator.info/docs/4.2/options

    :rtype: ColumnsGroup
    """
    return self.sub_data_enum("columns", ColumnsGroup)

  @property
  def columnVertAlign(self):
    """
    Description:
    -----------
    Vertical alignment for contents of column header (used in column grouping)

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["columnVertAlign"]

  @columnVertAlign.setter
  def columnVertAlign(self, val):
    self._attrs["columnVertAlign"] = val

  @property
  def data(self):
    """
    Description:
    -----------
    Array to hold data that should be loaded on table creation

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["data"]

  @data.setter
  def data(self, val):
    self._attrs["data"] = val

  @property
  def groupBy(self):
    """
    Description:
    -----------
    String/function to select field to group rows by

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["groupBy"]

  @groupBy.setter
  def groupBy(self, val):
    self._attrs["groupBy"] = val

  @property
  def groupToggleElement(self, val):
    """
    Description:
    -----------
    By default Tabulator allows users to toggle a group open or closed by clicking on the arrow icon in the left of the group header. If you would prefer a different behaviour you can use the groupToggleElement option to choose a different option:

    Related Pages:
http://tabulator.info/docs/4.0/group

    :param val:
    """
    self._attrs["groupToggleElement"] = val

  @groupToggleElement.setter
  def groupToggleElement(self, val):
    self._attrs["groupToggleElement"] = val

  @property
  def groupStartOpen(self):
    """
    Description:
    -----------
    You can set the default open state of groups using the groupStartOpen property.

    Related Pages:
http://tabulator.info/docs/4.0/group
    """
    return self._attrs["groupStartOpen"]

  @groupStartOpen.setter
  def groupStartOpen(self, val):
    self._attrs["groupStartOpen"] = val

  @property
  def groupValues(self):
    """
    Description:
    -----------
    Array of values for groups

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["groupValues"]

  @groupValues.setter
  def groupValues(self, val):
    self._attrs["groupValues"] = val

  @property
  def height(self):
    """
    Description:
    -----------
    Sets the height of the containing element, can be set to any valid height css value.
    If set to false (the default), the height of the table will resize to fit the table data.

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["height"]

  @height.setter
  def height(self, val):
    if isinstance(val, int):
      val = "%spx" % val
    self._attrs["height"] = val

  @property
  def history(self):
    """
    Description:
    -----------
    Enable user interaction history functionality

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["history"]

  @history.setter
  def history(self, val):
    self._attrs["history"] = val

  @property
  def lang(self):
    """
    Description:
    -----------
    hold localization templates

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["lang"]

  @lang.setter
  def lang(self, val):
    self._attrs["lang"] = val

  @property
  def layout(self):
    """
    Description:
    -----------
    Layout mode for the table columns

    Related Pages:
http://tabulator.info/docs/4.2/options

    :rtype: EnumLayout
    """
    return self.sub_data("layout", EnumLayout)

  @property
  def movableColumns(self):
    """
    Description:
    -----------
    Allow users to move and reorder columns

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["movableColumns"]

  @movableColumns.setter
  def movableColumns(self, val):
    self._attrs["movableColumns"] = val

  @property
  def movableRows(self):
    """
    Description:
    -----------
    Allow users to move and reorder rows

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["movableRows"]

  @movableRows.setter
  def movableRows(self, val):
    self._attrs["movableRows"] = val

  @property
  def movableRowsConnectedTables(self):
    """
    Description:
    -----------
    Connection selector for receving tables

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["movableRowsConnectedTables"]

  @movableRowsConnectedTables.setter
  def movableRowsConnectedTables(self, val):
    self._attrs["movableRowsConnectedTables"] = val

  @property
  def movableRowsReceiver(self):
    """
    Description:
    -----------
    Sender function to be executed when row has been received

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["movableRowsReceiver"]

  @movableRowsReceiver.setter
  def movableRowsReceiver(self, val):
    self._attrs["movableRowsReceiver"] = val

  @property
  def movableRowsSender(self):
    """
    Description:
    -----------
    Sender function to be executed when row has been sent

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["movableRowsSender"]

  @movableRowsSender.setter
  def movableRowsSender(self, val):
    self._attrs["movableRowsSender"] = val

  @property
  def pagination(self):
    """
    Description:
    -----------
    Choose pagination method, "local" or "remote"

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["pagination"]

  @pagination.setter
  def pagination(self, val):
    self._attrs["pagination"] = val

  @property
  def paginationSize(self):
    """
    Description:
    -----------
    Set the number of rows in each page

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["paginationSize"]

  @paginationSize.setter
  def paginationSize(self, val):
    if val is not None:
      #
      self._attrs["pagination"]= 'local'
    self._attrs["paginationSize"] = val

  @property
  def paginationSizeSelector(self):
    """
    Description:
    -----------
    Add page size selection select element to the table footer

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["paginationSizeSelector"]

  @paginationSizeSelector.setter
  def paginationSizeSelector(self, val):
    self._attrs["paginationSizeSelector"] = val

  @property
  def persistenceID(self):
    """
    Description:
    -----------
    ID tag used to identify persistent storage information

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["persistenceID"]

  @persistenceID.setter
  def persistenceID(self, val):
    self._attrs["persistenceID"] = val

  @property
  def persistence(self):
    """
    Description:
    -----------
    The persistence system has received an overhaul in this release, providing a more consistent way to configure table persistence and allow even more table options to be persisted between sessions.

    Related Pages:
http://tabulator.info/docs/4.5/release#persistence

    :rtype: Persistence
    """
    return self.sub_data("persistence", Persistence)

  @property
  def placeholder(self):
    """
    Description:
    -----------
    placeholder element to display on empty table

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["placeholder"]

  @placeholder.setter
  def placeholder(self, val):
    self._attrs["placeholder"] = val

  @property
  def printAsHtml(self):
    return self._attrs["printAsHtml"]

  @printAsHtml.setter
  def printAsHtml(self, val):
    self._attrs["printAsHtml"] = val

  @property
  def printHeader(self):
    return self._attrs["printHeader"]

  @printHeader.setter
  def printHeader(self, val):
    self._attrs["printHeader"] = val

  @property
  def printFooter(self):
    return self._attrs["printFooter"]

  @printFooter.setter
  def printFooter(self, val):
    self._attrs["printFooter"] = val

  @property
  def reactiveData(self):
    """
    Description:
    -----------
    enable data reactivity

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["reactiveData"]

  @reactiveData.setter
  def reactiveData(self, val):
    self._attrs["reactiveData"] = val

  @property
  def responsiveLayout(self):
    """
    Description:
    -----------
    Automatically hide/show columns to fit the width of the Tabulator element

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["responsiveLayout"]

  @responsiveLayout.setter
  def responsiveLayout(self, val):
    self._attrs["responsiveLayout"] = val

  @property
  def resizableColumns(self):
    """
    Description:
    -----------
    Allow user to resize columns (via handles on the left and right edges of the column header)

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["resizableColumns"]

  @resizableColumns.setter
  def resizableColumns(self, val):
    self._attrs["resizableColumns"] = val

  @property
  def selectable(self):
    """
    Description:
    -----------
    Enable/Disable row selection

    Related Pages:
http://tabulator.info/docs/4.2/options
    """
    return self._attrs["selectable"]

  @selectable.setter
  def selectable(self, val):
    self._attrs["selectable"] = val


class TableTreeConfig(TableConfig):

  @property
  def dataTree(self):
    return self._attrs["dataTree"]

  @dataTree.setter
  def dataTree(self, val):
    self._attrs["dataTree"] = val

  @property
  def dataTreeStartExpanded(self):
    return self._attrs["dataTreeStartExpanded"]

  @dataTreeStartExpanded.setter
  def dataTreeStartExpanded(self, val):
    self._attrs["dataTreeStartExpanded"] = val



"""
Module dedicated to secured data

It will rely on the external package cryptography
"""

import os
import base64
import uuid

# Internal key used for the token generation in the URL (during the sharing)
KEY = 'bAvGUC_7oazo4cIeNBt8t23bPe3Hvq8livGqQxSV-z0='


class PyCrypto(object):
  class __internal(object):
    _props, _context = {}, {}

  def __init__(self, src=None):
    self.__src = src if src else self.__internal()

  def encrypt(self, data, token=None, salt=None):
    """
    This function will use the cryptography to ensure a secured encryption of the different credential and private data.
    This can be also used to protect data from the report.
    In order to ensure the right privacy please do not store the token and the salt in the framework.

    example

    Documentation

    :param data: The data to be encrypted
    :param token: Optional. The token used to encrypt the data
    :param salt: Optional. The salt id

    :return: The encrypted data with the salt used
    """
    from cryptography.fernet import Fernet
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import hashes
    from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

    if token is None:
      # used for the URL generation
      return Fernet(KEY).encrypt(data.encode())

    salt = salt if salt is not None else os.urandom(32)
    kdf = PBKDF2HMAC(algorithm=hashes.SHA256(), length=32, salt=salt, iterations=100000, backend=default_backend())
    if hasattr(token, 'encode'):
      encrypted_key = base64.urlsafe_b64encode(kdf.derive(bytes(token.encode('utf-8'))))
    else:
      encrypted_key = token
    return Fernet(encrypted_key).encrypt(bytes(data.encode('latin1'))).decode('latin1'), salt.decode('latin1')

  def decrypt(self, encrypted, token=None, salt=None, label=''):
    """
    This function will use the two keys in order to decrypt the data.
    In case of failure this function will raise an exception.

    Example
    PyCrypto().decrypt(encrypted)

    Documentation

    :param encrypted: The encrypted data
    :param token: Optional. The token used to encrypt the data
    :param salt: Optional. The salt id
    :param label: Optional. A label used to store the reference in the log file

    :return: A string with the decrypted data
    """
    from cryptography.fernet import Fernet
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import hashes
    from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

    if token is None:
      # used for the URL generation
      return Fernet(KEY).decrypt(encrypted)

    if hasattr(self.__src, 'log'):
      # In the admin section the report is not defined
      self.__src.log("SECURITY|%s|%s|password decoding" % (self.__src.run.current_user, label))
    kdf = PBKDF2HMAC(algorithm=hashes.SHA256(), length=32, salt=bytes(salt.encode('latin1')), iterations=100000, backend=default_backend())
    if hasattr(token, 'encode'):
      encrypted_key = base64.urlsafe_b64encode(kdf.derive(bytes(token.encode('utf-8'))))
    else:
      encrypted_key = token
    return Fernet(encrypted_key).decrypt(bytes(encrypted.encode('latin1'))).decode('utf-8')

  @property
  def getId(self):
    """
    Return a unique token

    Example
    PyCrypto.getId

    Related Pages:

			https://docs.python.org/2/library/uuid.html

    :return: A unique ID (based on the timestamp
    """
    return uuid.uuid4()

  @property
  def key(self):
    """
    Return a Fernet key

    Example
    PyCrypto().key

    Related Pages:

			https://cryptography.io/en/latest/fernet/
    """
    from cryptography.fernet import Fernet
    return Fernet.generate_key()

  @classmethod
  def b64encode(cls, text, salt=None):
    """

    Example
    PyCrypto.b64encode("Test")

    Related Pages:

			https://cryptography.io/en/latest/hazmat/primitives/key-derivation-functions/

    :param text: The text to be encrypted
    :param salt: Optional. The salt used for the encryption (default None)

    :return:
    """
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import hashes
    from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

    salt = salt if salt is not None else os.urandom(32)
    kdf = PBKDF2HMAC(algorithm=hashes.SHA256(), length=32, salt=salt, iterations=100000, backend=default_backend())
    return base64.urlsafe_b64encode(kdf.derive(bytes(text.encode('latin1'))))

  @classmethod
  def cryptKeyPairs(cls, msg, key1, key2):
    """

    Example

    Documentation

    :param msg:
    :param key1:
    :param key2:
    :return:
    """
    from cryptography.fernet import Fernet, MultiFernet
    f = MultiFernet([Fernet(key1), Fernet(key2)])
    return f.encrypt(bytes(msg.encode('latin1')))

  @classmethod
  def decryptKeyPairs(cls, encrypted, key1, key2):
    """

    Example

    Documentation

    :param encrypted: The encrypted data
    :param key1:
    :param key2:
    :return:
    """
    from cryptography.fernet import Fernet, MultiFernet
    f = MultiFernet([Fernet(key1), Fernet(key2)])
    return f.decrypt(bytes(encrypted.encode('latin1')))






from epyk.core import html
from epyk.core.html import graph

from epyk.core.html import Defaults as defaults_html
from epyk.core.css import Defaults as defaults_css


class Numbers(object):

  def __init__(self, context):
    self.context = context

  def digits(self, text=None, color=None, align='center', width=None, height=None, htmlCode=None,
           tooltip=None, options=None, profile=None):
    """
    Description:
    ------------
    The <span> tag is used to group inline-elements in a document.

    The <span> tag provides no visual change by itself.

    The <span> tag provides a way to add a hook to a part of a text or a part of a document.

    Usage::

      rptObj.ui.texts.span("Test")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlText.Position`

    Related Pages:

        https://www.w3schools.com/tags/tag_span.asp

    Attributes:
    ----------
    :param text: Optional. The string value to be displayed in the component
    :param color: Optional. The color of the text
    :param align: Optional. The position of the icon in the line (left, right, center)
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param htmlCode: Optional. An identifier for this component (on both Python and Javascript side)
    :param tooltip: Optional. A string with the value of the tooltip
    :param profile: Optional. A flag to set the component performance storage
    """
    if width is None:
      width = (defaults_html.TEXTS_SPAN_WIDTH, 'px')
    if height is None:
      height = (defaults_html.LINE_HEIGHT, 'px')
    html_label = html.HtmlText.Position(self.context.rptObj, text, color, align, width, height, htmlCode, tooltip, options, profile)
    html_label.position(3, {"font-size": defaults_css.font(5), "font-weight": "bold"})
    html_label.position(4, {"font-size": defaults_css.font(5), "font-weight": "bold"})
    html_label.digits(True)
    return html_label

  def number(self, number=0, title=None, label=None, icon=None, color=None, tooltip='', htmlCode=None,
             options=None, helper=None, width=(150, 'px'), profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.texts.number(289839898, label="test", helper="Ok", icon="fas fa-align-center")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlText.Numeric`

    Attributes:
    ----------
    :param number: Optional. The value to be displayed to the component. Default now
    :param title:
    :param label: Optional. The text of label to be added to the component
    :param icon: Optional. A string with the value of the icon to display from font-awesome
    :param color:
    :param tooltip:
    :param htmlCode:
    :param options:
    :param helper:
    :param profile:
    """
    dflt_options = {"digits": 0, "thousand_sep": ',', "decimal_sep": '.'}
    if options is not None:
      dflt_options.update(options)
    html_number = html.HtmlText.Numeric(self.context.rptObj, number, title, label, icon, color, tooltip, htmlCode,
                                        dflt_options, helper, width, profile)
    return html_number

  def pound(self, number=0, title=None, label=None, icon=None, color=None, tooltip='', htmlCode=None,
             options=None, helper=None, width=(150, 'px'), profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.texts.pound(289839898, label="test", helper="Ok", icon="fas fa-align-center")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlText.Numeric`

    Attributes:
    ----------
    :param number: Optional. The value to be displayed to the component. Default now
    :param title:
    :param label: Optional. The text of label to be added to the component
    :param icon: Optional. A string with the value of the icon to display from font-awesome
    :param color:
    :param tooltip:
    :param htmlCode:
    :param options:
    :param helper:
    :param profile:
    """
    html_number = self.number(number, title, label, icon, color, tooltip, htmlCode, options, helper, width, profile)
    html_number.money("£")
    return html_number

  def euro(self, number=0, title=None, label=None, icon=None, color=None, tooltip='', htmlCode=None,
             options=None, helper=None, width=(150, 'px'), profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.texts.euro(289839898, label="test", helper="Ok", icon="fas fa-align-center")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlText.Numeric`

    Attributes:
    ----------
    :param number: Optional. The value to be displayed to the component. Default now
    :param title:
    :param label: Optional. The text of label to be added to the component
    :param icon: Optional. A string with the value of the icon to display from font-awesome
    :param color:
    :param tooltip:
    :param htmlCode:
    :param options:
    :param helper:
    :param profile:
    """
    html_number = self.number(number, title, label, icon, color, tooltip, htmlCode, options, helper, width, profile)
    html_number.money("€", format="%v %s")
    return html_number

  def dollar(self, number=0, title=None, label=None, icon=None, color=None, tooltip='', htmlCode=None,
             options=None, helper=None, width=(150, 'px'), profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.texts.dollar(289839898, label="test", helper="Ok", icon="fas fa-align-center")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlText.Numeric`

    Attributes:
    ----------
    :param number: Optional. The value to be displayed to the component. Default now
    :param title:
    :param label: Optional. The text of label to be added to the component
    :param icon: Optional. A string with the value of the icon to display from font-awesome
    :param color:
    :param tooltip:
    :param htmlCode:
    :param options:
    :param helper:
    :param profile:
    """
    html_number = self.number(number, title, label, icon, color, tooltip, htmlCode, options, helper, width, profile)
    html_number.money("$", format="%v %s")
    return html_number

  def money(self, symbol, number=0, title=None, label=None, icon=None, color=None, tooltip='', htmlCode=None,
             options=None, helper=None, width=(150, 'px'), profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.texts.money(289839898, label="test", helper="Ok", icon="fas fa-align-center")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlText.Numeric`

    Attributes:
    ----------
    :param number: Optional. The value to be displayed to the component. Default now
    :param title:
    :param label: Optional. The text of label to be added to the component
    :param icon: Optional. A string with the value of the icon to display from font-awesome
    :param color:
    :param tooltip:
    :param htmlCode:
    :param options:
    :param helper:
    :param profile:
    """
    html_number = self.number(number, title, label, icon, color, tooltip, htmlCode, options, helper, width, profile)
    html_number.money(symbol, format="%v %s")
    return html_number

  def plotly(self, value, profile=None, options=None, width=(100, "%"), height=(330, "px"),
             htmlCode=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.graph.GraphPlotly.Indicator`

    :param value:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    :return:
    """
    ind = graph.GraphPlotly.Indicator(self.context.rptObj, width, height, options or {}, htmlCode, profile)
    ind.add_trace({'value': value}, mode="number")
    return ind

  def plotly_with_delta(self, value, profile=None, options=None, width=(100, "%"),
                        height=(330, "px"), htmlCode=None):
    """
    Underlying HTML Objects:

      - :class:`epyk.core.graph.GraphPlotly.Indicator`

    :param value:
    :param profile:
    :param options:
    :param width:
    :param height:
    :param htmlCode:
    :return:
    """
    ind = graph.GraphPlotly.Indicator(self.context.rptObj, width, height, options or {}, htmlCode, profile)
    ind.add_trace({'value': value}, mode="number+delta")
    return ind



from epyk.core import html
from epyk.core.css import Colors


class Images(object):
  def __init__(self, context):
    self.context = context

  def img(self, image=None, path=None, width=(100, "%"), height=(None, "px"), align="center", htmlCode=None,
          profile=None, options=None):
    """
    Description:
    ------------
    Add an HTML image to the page. The path can be defined either in a absolute or relative format.

    Tip: The absolute format does not work on servers. It is recommended to use relative starting to the root of the server

    Usage::

      rptObj.ui.img("epykIcon.PNG", path=r"../../../static/images", height=(50, "px"))


    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlImage.Image`

    Related Pages:

      https://www.w3schools.com/bootstrap/bootstrap_ref_css_images.asp
      https://www.w3schools.com/cssref/css3_pr_border-radius.asp

    Attributes:
    ----------
    :param image: String. The image file name
    :param path: Optional. String. The image file path
    :param width: Optional. Tuple. The component width in pixel or percentage
    :param height: Optional. Tuple. The component height in pixel or percentage
    :param align:
    :param htmlCode:
    :param profile:
    :param options:
    """
    if height[0] is not None and width[1] == '%':
      width = ("auto", '')
    html_image = html.HtmlImage.Image(self.context.rptObj, image, path, align, htmlCode, width, height, profile, options or {})
    return html_image

  def circular(self, image=None, path=None, width=(100, "%"), height=('auto', ""), align="center", htmlCode=None,
          profile=None, options=None):
    """
    Description:
    ------------
    Add an HTML image to the page. The path can be defined either in a absolute or relative format.

    Tip: The absolute format does not work on servers. It is recommended to use relative starting to the root of the server

    Usage::

      rptObj.ui.circular("epykIcon.PNG", path=r"../../../static/images", height=(50, "px"))

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlImage.Image`

    Related Pages:

			https://www.w3schools.com/bootstrap/bootstrap_ref_css_images.asp
    https://www.w3schools.com/cssref/css3_pr_border-radius.asp

    Attributes:
    ----------
    :param image: String. The image file name
    :param path: Optional. String. The image file path
    :param width: Optional. Tuple. The component width in pixel or percentage
    :param height: Optional. Tuple. The component height in pixel or percentage
    :param align:
    :param htmlCode:
    :param profile:
    :param options:
    """
    if height[0] is not None and width[1] == '%':
      width = ("auto", '')
    html_image = html.HtmlImage.Image(self.context.rptObj, image, path, align, htmlCode, width, height, profile, options or {})
    # add the css styles
    html_image.style.css.padding = 5
    html_image.style.css.borders_light()
    html_image.style.css.border_radius = 50
    return html_image

  def avatar(self, text=None, image=None, path=None, status=None, width=(30, "px"), height=(30, "px"), align="center", htmlCode=None,
               profile=None, options=None):
    """
    Description:
    ------------
    Generate or load an avatar

    Usage::

      rptObj.ui.images.avatar("Epyk", status='out')
      rptObj.ui.images.avatar(image="epykIcon.PNG", path=config.IMG_PATH, status=False)

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Div`
      - :class:`epyk.core.html.HtmlImage.Image`

    Attributes:
    ----------
    :param image:
    :param path:
    :param width:
    :param height:
    :param align:
    :param htmlCode:
    :param profile:
    :param options:
    """
    status_map = {
      True: self.context.rptObj.theme.success[1],
      'available': self.context.rptObj.theme.success[1],
      False: self.context.rptObj.theme.danger[1],
      'busy': self.context.rptObj.theme.danger[1],
      'out': self.context.rptObj.theme.warning[1]
    }

    bgcolor, margin_top = None, -5
    if image is not None:
      img = self.img(image, path, (width[0]-5, width[1]), (height[0]-5, height[1]), align, htmlCode, profile, options)
      img.style.css.border_radius = 50
      img.style.css.margin = 2
      margin_top = -8
    else:
      bgcolor = Colors.randColor(self.context.rptObj.py.hash(text))
      img = self.context.rptObj.ui.layouts.div(text[0].upper())
      img.style.css.line_height = width[0] - 5
      img.style.css.color = "white"
      img.style.css.font_size = width[0]
      img.style.css.font_weight = 'bold'
      img.style.css.padding = 0
    img.style.css.middle()
    status_o = self.context.rptObj.ui.layouts.div("&nbsp;", width=(10, "px"), height=(10, "px"))
    status_o.style.css.position = 'relative'

    status_o.style.css.background_color = status_map.get(status, self.context.rptObj.theme.greys[5])
    status_o.style.css.border_radius = 10
    status_o.style.css.margin_top = margin_top
    status_o.style.css.float = 'right'

    div = self.context.rptObj.ui.layouts.div([img, status_o], width=width, height=height)
    if bgcolor is not None:
      div.style.css.background_color = bgcolor
      div.style.css.text_stoke = "1px %s" % bgcolor
    div.style.css.borders_light()
    div.style.css.border_radius = 50
    div.img = img
    div.status = status_o
    return div

  def section(self, image, name, title, text, url=None, path=None, width=(200, "px"), height=(200, "px")):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.images.section("epykIcon.PNG", "# Test", "Epyk Test", 'This is a test', path=r"../../../static/images")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Div`
      - :class:`epyk.core.html.HtmlText.Span`
      - :class:`epyk.core.html.HtmlText.Paragraph`
      - :class:`epyk.core.html.HtmlText.Title`
      - :class:`epyk.core.html.HtmlImage.Image`

    Attributes:
    ----------
    :param image:
    :param name:
    :param title:
    :param text:
    :param url:
    :param path:
    :param width:
    :param height:
    """
    img = self.img(image, width=(width[0]-10, 'px'), height=(100, "px"), path=path)
    title = self.context.rptObj.ui.title(title, level=2)
    highlight = self.context.rptObj.ui.texts.span(name, width=(50, "px"), height=(20, 'px'))
    paragraph = self.context.rptObj.ui.texts.paragraph(text)
    div = self.context.rptObj.ui.layouts.div([highlight, img, title, paragraph], width=width, height=height)
    highlight.css({"position": "absolute", 'left': 0, "background-color": self.context.rptObj.theme.colors[-1],
                   "color": self.context.rptObj.theme.greys[0], 'padding': "0 2px"})
    div.style.css.margin = 2
    if url is not None:
      div.style.css.cursor = 'pointer'
      div.click([self.context.rptObj.js.location.href(url)])
    div.style.add_classes.div.border_bottom()
    return div

  def animated(self, image=None, text="", title="", url=None, path=None, width=(200, "px"), height=(200, "px"),
               profile=None):
    """
    Description:
    ------------
    Advance image with mask and gallery link.
    This will display some details when the mouse is on the container

    Usage::

      c = rptObj.ui.images.animated("epykIcon.PNG", text="This is a comment", title="Title", url="#", path=r"../../../static/images")
      c.style.css.borders()

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlImage.AnimatedImage`

    Related Pages:

			https://tympanus.net/Tutorials/OriginalHoverEffects/

    Attributes:
    ----------
    :param image: String. The image file name
    :param text: Optional. String. The image file path
    :param title: String. The image title displayed in to the mask on mouse hover the container
    :param url: String. The link to the gallery
    :param path: Optional. String. The image file path
    :param width: Optional. Tuple. The component width in pixel or percentage
    :param height: Optional. Tuple. The component height in pixel or percentage
    :param profile:
    """
    html_id = html.HtmlImage.AnimatedImage(self.context.rptObj, image, text, title, url, path, width, height, profile)
    return html_id

  def carrousel(self, images, path=None, selected=0, width=(100, "%"), height=(300, "px"), profile=None):
    """
    Description:
    ------------
    Carousel component for pictures

    Usage::

      car = rptObj.ui.images.carrousel(["epykIcon.PNG", "epyklogo.ico", "epyklogo_whole_big.png"],
                                 path=r"../../../static/images", height=(200, 'px'))
      car.click([rptObj.js.console.log('data', skip_data_convert=True)])

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlImage.ImgCarrousel`

    Related Pages:

			https://www.cssscript.com/basic-pure-css-slideshow-carousel/

    Attributes:
    ----------
    :param images: List. With the different picture file names
    :param path: String. The common path for the pictures
    :param width: Optional. Tuple. The component width in pixel or percentage
    :param height: Optional. Tuple. The component height in pixel
    :param profile:
    """
    if height[1] == '%':
      raise Exception("This height cannot be in percentage")

    html_i = html.HtmlImage.ImgCarrousel(self.context.rptObj, images, path, selected, width, height, profile)
    return html_i

  def emoji(self, symbole=None, top=(20, 'px'), profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.images.emoji(rptObj.symbols.smileys.DISAPPOINTED_FACE)

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlImage.Emoji`

    Related Pages:

			https://github.com/wedgies/jquery-emoji-picker

    Attributes:
    ----------
    :param symbole:
    :param top:
    :param profile:
    """
    html_emoji = html.HtmlImage.Emoji(self.context.rptObj, symbole, top, profile)
    return html_emoji

  def icon(self, icon=None, width=(None, 'px'), height=(None, "px"), color=None, tooltip=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.images.icon("fab fa-angellist")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlImage.Icon`

    Related Pages:

			https://fontawesome.com/icons?m=free

    Attributes:
    ----------
    :param icon:
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param color:
    :param tooltip:
    :param profile:
    """
    html_icon = html.HtmlImage.Icon(self.context.rptObj, icon, width=width, height=height, color=color, tooltip=tooltip,
                                    profile=profile)
    return html_icon

  def badge(self, text="", label=None, icon=None, width=(25, "px"), height=(25, "px"), background_color=None, color=None, url=None,
            tooltip=None, options=None, profile=None):
    """
    Description:
    ------------
    Display a badge component using Bootstrap

    Usage::

      rptObj.ui.images.badge("Test badge", "Label", icon="fas fa-align-center")
      rptObj.ui.images.badge("This is a badge", background_color="red", color="white")
      rptObj.ui.images.badge(12, icon="far fa-bell", options={"badge_position": 'right'})

      b = rptObj.ui.images.badge(7688, icon="fab fa-python", options={'badge_css': {'color': 'white', "background": 'red'}})
      b.options.badge_css = {"background": 'green'}

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlImage.Badge`

    Related Pages:

        https://getbootstrap.com/docs/4.0/components/badge/

    Attributes:
    ----------
    :param text: The content of the badge
    :param label: Optional, The label to display close to the badge
    :param icon: Optional, A String with the icon to display from font-awesome
    :param background_color: Optional, The background color of the badge
    :param color: Optional, The text color of the badge
    :param url:
    :param tooltip: Optional, The text to display in the tooltip
    :param options:
    :param profile: Optional, A boolean to store the performances for each components
    """
    if background_color is None:
      background_color = self.context.rptObj.theme.greys[0]
    if color is None:
      color = self.context.rptObj.theme.success[1]
    html_badge = html.HtmlImage.Badge(self.context.rptObj, text, width, height, label, icon, background_color, color, url,
                                      tooltip, options or {}, profile)
    return html_badge

  def color(self, code, color=None, width=(110, 'px'), height=(25, 'px'), options=None, helper=None, profile=None):
    """
    Description:
    ------------
    Simple vignette to display a color with it is code

     Usage::

      rptObj.ui.images.color("FFFFFF")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlContainer.Div`

    Attributes:
    ----------
    :param code: String. The color code
    :param color: String. The font color
    :param width: Tuple. The size with its unit
    :param height: Tuple. The size with its unit
    :param options: Dictionary. The object options
    :param helper:
    :param profile:
    """
    div = self.context.rptObj.ui.div(code, width=width, height=height, options=options, helper=helper, profile=profile)
    div.style.css.background_color = code
    div.style.css.line_height = "%s%s" % (height[0], height[1])
    div.style.css.color = color or self.context.rptObj.theme.greys[0]
    div.style.css.text_align = "center"
    div.style.css.border = "1px solid black"
    div.style.css.vertical_align = "middle"
    return div



"""
Related Pages:

		https://www.w3schools.com/charsets/ref_utf_punctuation.asp
"""

from epyk.core.html.symboles import SymbSmileys
from epyk.core.html.symboles import SymbHtml
from epyk.core.html.symboles import SymbArrows
from epyk.core.html.symboles import SymbCurrencies
from epyk.core.html.symboles import SymbLetters
from epyk.core.html.symboles import SymbEmojis
from epyk.core.html.symboles import SymbMaths
from epyk.core.html.symboles import SymbPunctuations
from epyk.core.html.symboles import SymbShapes


class Symboles(object):
  @property
  def shapes(self):
    """

    Related Pages:

			https://www.w3schools.com/charsets/ref_utf_geometric.asp
    """
    return SymbShapes

  @property
  def punctuations(self):
    """

    Related Pages:

			https://www.w3schools.com/charsets/ref_utf_punctuation.asp
    """
    return SymbPunctuations

  @property
  def maths(self):
    """

    Related Pages:

			https://www.w3schools.com/charsets/ref_utf_math.asp
    """
    return SymbMaths

  @property
  def emojis(self):
    """

    Related Pages:

			https://www.w3schools.com/charsets/ref_emoji.asp
    """
    return SymbEmojis

  @property
  def letters(self):
    """

    Related Pages:

			https://www.w3schools.com/charsets/ref_utf_letterlike.asp
    """
    return SymbLetters

  @property
  def currencies(self):
    """

    Related Pages:

			https://www.w3schools.com/charsets/ref_utf_currency.asp
    """
    return SymbCurrencies

  @property
  def smileys(self):
    """

    Related Pages:

			https://www.w3schools.com/charsets/ref_emoji_smileys.asp
    """
    return SymbSmileys

  @property
  def html(self):
    """

    Related Pages:

			https://www.w3schools.com/html/html_symbols.asp
    https://www.w3schools.com/charsets/ref_utf_math.asp
    """
    return SymbHtml

  @property
  def arrows(self):
    """

    Related Pages:

			https://www.w3schools.com/charsets/ref_utf_arrows.asp
    """
    return SymbArrows


if __name__ == '__main__':
  data = '''
†	dagger	02020	8224
ℸ	daleth	02138	8504
↡	Darr	021A1	8609
⇓	dArr	021D3	8659
↓	darr	02193	8595
‐	dash	02010	8208
⫤	Dashv	02AE4	10980
⊣	dashv	022A3	8867
⤏	dbkarow	0290F	10511
˝	dblac	002DD	733
Ď	Dcaron	0010E	270
ď	dcaron	0010F	271
Д	Dcy	00414	1044
д	dcy	00434	1076
ⅅ	DD	02145	8517
ⅆ	dd	02146	8518
‡	ddagger	02021	8225
⇊	ddarr	021CA	8650
⤑	DDotrahd	02911	10513
⩷	ddotseq	02A77	10871
°	deg	000B0	176
∇	Del	02207	8711
Δ	Delta	00394	916
δ	delta	003B4	948
⦱	demptyv	029B1	10673
⥿	dfisht	0297F	10623
𝔇	Dfr	1D507	120071
𝔡	dfr	1D521	120097
⥥	dHar	02965	10597
⇃	dharl	021C3	8643
⇂	dharr	021C2	8642
´	DiacriticalAcute	000B4	180
˙	DiacriticalDot	002D9	729
˝	DiacriticalDoubleAcute	002DD	733
`	DiacriticalGrave	00060	96
˜	DiacriticalTilde	002DC	732
⋄	diam	022C4	8900
⋄	Diamond	022C4	8900
⋄	diamond	022C4	8900
♦	diamondsuit	02666	9830
♦	diams	02666	9830
¨	die	000A8	168
ⅆ	DifferentialD	02146	8518
ϝ	digamma	003DD	989
⋲	disin	022F2	8946
÷	div	000F7	247
÷	divide	000F7	247
⋇	divideontimes	022C7	8903
⋇	divonx	022C7	8903
Ђ	DJcy	00402	1026
ђ	djcy	00452	1106
⌞	dlcorn	0231E	8990
⌍	dlcrop	0230D	8973
$	dollar	00024	36
𝔻	Dopf	1D53B	120123
𝕕	dopf	1D555	120149
¨	Dot	000A8	168
˙	dot	002D9	729
⃜	DotDot	020DC	8412
≐	doteq	02250	8784
≑	doteqdot	02251	8785
≐	DotEqual	02250	8784
∸	dotminus	02238	8760
∔	dotplus	02214	8724
⊡	dotsquare	022A1	8865
⌆	doublebarwedge	02306	8966
∯	DoubleContourIntegral	0222F	8751
¨	DoubleDot	000A8	168
⇓	DoubleDownArrow	021D3	8659
⇐	DoubleLeftArrow	021D0	8656
⇔	DoubleLeftRightArrow	021D4	8660
⫤	DoubleLeftTee	02AE4	10980
⟸	DoubleLongLeftArrow	027F8	10232
⟺	DoubleLongLeftRightArrow	027FA	10234
⟹	DoubleLongRightArrow	027F9	10233
⇒	DoubleRightArrow	021D2	8658
⊨	DoubleRightTee	022A8	8872
⇑	DoubleUpArrow	021D1	8657
⇕	DoubleUpDownArrow	021D5	8661
∥	DoubleVerticalBar	02225	8741
↓	DownArrow	02193	8595
⇓	Downarrow	021D3	8659
↓	downarrow	02193	8595
⤓	DownArrowBar	02913	10515
⇵	DownArrowUpArrow	021F5	8693
̑	DownBreve	00311	785
⇊	downdownarrows	021CA	8650
⇃	downharpoonleft	021C3	8643
⇂	downharpoonright	021C2	8642
⥐	DownLeftRightVector	02950	10576
⥞	DownLeftTeeVector	0295E	10590
↽	DownLeftVector	021BD	8637
⥖	DownLeftVectorBar	02956	10582
⥟	DownRightTeeVector	0295F	10591
⇁	DownRightVector	021C1	8641
⥗	DownRightVectorBar	02957	10583
⊤	DownTee	022A4	8868
↧	DownTeeArrow	021A7	8615
⤐	drbkarow	02910	10512
⌟	drcorn	0231F	8991
⌌	drcrop	0230C	8972
𝒟	Dscr	1D49F	119967
𝒹	dscr	1D4B9	119993
Ѕ	DScy	00405	1029
ѕ	dscy	00455	1109
⧶	dsol	029F6	10742
Đ	Dstrok	00110	272
đ	dstrok	00111	273
⋱	dtdot	022F1	8945
▿	dtri	025BF	9663
▾	dtrif	025BE	9662
⇵	duarr	021F5	8693
⥯	duhar	0296F	10607
⦦	dwangle	029A6	10662
Џ	DZcy	0040F	1039
џ	dzcy	0045F	1119
⟿	dzigrarr	027FF	10239
  '''

  for rec in data.split("\n"):
    split_line = rec.split("\t")
    if len(split_line) == 4:
      code = "&#%s" % split_line[3] if not split_line[3].startswith("&#") else split_line[3]
      print("%s = '%s'" % (split_line[1].strip().replace(" ", "_").replace("-", "_"), code))
    if len(split_line) > 4 and split_line[4].strip() != "":
      if split_line[1].endswith(";"):
        code = split_line[1]
      else:
        code = "&#%s" % split_line[1].strip() if not split_line[1].startswith("&#") else split_line[1].strip()
      if code == "&#" or code == '':
        continue

      names = split_line[4].split("=")
      for name in names:
        print("%s = '%s'" % (name.strip().replace(" ", "_").replace("-", "_").upper(), code))
